unit DataFile;
interface

{$I options.pp}

uses
 {$IFDEF Turbo} Skl_Defs, {$ENDIF}
 {$IFDEF FPC  } Sk_types, {$ENDIF}
 Disk_Rec, Records, Views;

const
 FileLinkMax = 20;
 StdIndxBufLen:byte = 64;
 StdDataBufLen:byte = 64;

type
 LongFName = string[12];
 ShortFName = string[4];

 TVirtualList = object
  Next:pointer; Size:word;
  constructor Init(aSize:word);
  function TestEl(var Rec,Patt):boolean; virtual;
  function Test(var Rec):boolean; virtual;
  function Add(var Rec):boolean; virtual;
  function Del(var Rec):boolean; virtual;
 end;
 TRecProtList = object (TVirtualList)
  constructor Init;
  function TestEl(var Rec,Patt):boolean; virtual;
 end;

 TFieldId = record FileId, FldId:byte end;
 PIndex = ^TIndex;

 PDataFile = ^TDataFile;
 TDataFile = object
  FName:SmallFName;
  LName:LongFName;
  SName:ShortFName;
  Status:byte; { 0 - norm, 1 - temp }
  CntSize:byte; { Column for rec counter }
  FLCnt:byte;
  FLArr:array[0..FileLinkMax] of TFieldId;
  LastFileLink, LastUsedIndex:byte;
  RecSize:word;
  Data:PRecFileCashDrv;
  Fields:PRecFileCashDrv;
  PLessP:pointer; CLessP:word;
  IndArr:array[0..FldMaxCnt] of PIndex;
  RecProtList:TRecProtList;
  OpenLen:longint;
  PntrSummExists:boolean;
  constructor Open(var DFNR);
  constructor Create(var DFNR);
  destructor Kill;
  procedure InitDataFile;
  procedure InitFieldsFile;
  procedure MakeCopy(FN:SmallFName);
  function IndexName(FldId:byte):string;
  function FldName(FldId:byte):string;
  function FldNameLnks(FldId:byte):string;
  function LinkChois:byte;
  function PntrChois:byte;
  function FldType(FldId:byte):byte;
  function FldULin(FldId:byte):boolean;
  function FldPos(FldId:byte):word;
  function RelationsCount:byte;
  function FullShowCount:byte;
  procedure SelectIndex(var IndexId:byte);
  function SelectFIndex:byte;
  function FindFirst(FldId:byte; var R; Dir:boolean):longint;
  procedure CreateIndex(FldId:byte);
  procedure EraseIndex(FldId:byte);
  procedure OpenIndex;
  procedure ExtrIndData(
   FldId:byte; TmpIName:string; var TmpISize:word; PV:PView);
  function Less(p1,p2:longint; FldId:byte):boolean; virtual;
  function LessP(p:longint; FldId:byte; Dir:boolean):boolean; virtual;
  procedure AppendField(FldId:byte; var Fld:TFldRec);
  procedure ChangeField(FldId:byte; var OFld,NFld:TFldRec);
  procedure EraseField(FldNum:byte);
  procedure SwapField(FldNum:byte);
  procedure SetFldLinks;
  function StdHeaderLine(ExFld:byte; var SummLineExist:boolean):string;
  function StdSummLine(ExFld:byte):string;
  function GetCount(Ind:longint; FileId,FldId:byte):longint;
  function GetPSumm(Ind:longint; FldId:byte):string;
  function StdBrouseLine(
   Ind,Num:longint; var Rec; ExFld:byte; SummLine:boolean):string;
  function StdPtrLine(p:longint; Depth:byte):string;
  procedure RecClear(var Rec);
  procedure TestForUnique(ExI:Longint; var NewR; var SucInd:longint);
  procedure RecUpdate(var Ind:longint; var NewR; old,always:boolean);
  function ExchRelated(Ind,Dest:longint):boolean;
  function RecErase(Ind,Dest:longint):boolean;
 end;
 TSystemMode = (InitMode, UseMode, DesignMode);

{------------------ Index driver: ---------------------}

 TValue = longint;
 TPlace = longint;
 TIPart = record
  Cnt:longint;
  Ptr:longint;
 end;
 TINode = record
  V:TValue;
  L,R:TIPart;
  B:longint;
 end;
 TIndex = object (TRecFileCashDrv)
  Data:TIPart;
  PDF:PDataFile;
  FldId:byte;
  constructor Init(aName:FullFName; aPDF:PDataFile; aFldId:byte);
  constructor Build(
   aName:FullFName; aPDF:PDataFile; aFldId:byte;
   TmpIName:FullFName; TmpISize:word; PV:PView);
  procedure ChangeName(aName:FullFName; aFldId:byte);
  procedure Add(V:TValue);
  procedure Del(V:TValue);
  procedure Mov(OV,NV:TValue);
  function Find(aV:TValue; var OK:boolean):TPlace;
  function FindFirst(Dir:boolean):TPlace;
  function GetV(V:TPlace):TValue;
{  destructor Done; }
  destructor Kill;
  procedure Show(P:TIPart);
 private
  N:longint;
  function Less(X,Y:TValue):boolean;
  procedure AddPart(var P:TIPart; Root:longint; aV:TValue);
  procedure DelPart(var P:TIPart; aV:TValue);
  function  GetVPart(P:TIPart; aV:TPlace):TValue;
  procedure ChB(P:TIPart; Root:longint);
  procedure GclAdd(Pt:longint);
  procedure Garbage;
  procedure IndFlush;
 end;

{
procedure TIPartClear(var TIP:TIPart);
procedure TINodeClear(var TIN:TINode; aV:TValue; Root:longint);
}

{------------------------------------------------------}

const
 Root = $FFFFFFFF; { for Index Nil value }

function DF_FindId(FN:SmallFName):byte;
function GetFileId(FLCursor:longint):byte;

procedure SetUseMode;
procedure SetDesignMode;
procedure ClearFileLinks;
procedure AddFileLink(FLSors,FLDest,FieldId:byte);

var
 SystemMode:TSystemMode;
 DFArr:array[0..254] of PDataFile;

implementation

uses
 Trace,
 File_win,
 TV_box, TV_Sort, Win_Def, Menu_cmd, Info_def, Str_Def,
 App, Drivers, Objects;

{------------------------------------------------------}

constructor TVirtualList.Init(aSize:word);
 begin Size:=aSize; Next:=Nil end;
function TVirtualList.TestEl(var Rec,Patt):boolean; begin abstract end;
function TVirtualList.Test(var Rec):boolean;
 var P:^pointer; D:PBlock absolute P; OK:boolean;
 begin
  P:=Next; OK:=False;
  while (not OK) and (P<>nil) do
   begin OK:=TestEl(Rec,D^[SizeOf(Next)]); P:=P^ end;
  Test:=OK;
 end;
function TVirtualList.Add(var Rec):boolean;
 var P:PBlock;
 begin if Test(Rec)
  then Add:=False
  else begin
   GetMem(P,Size+SizeOf(Next));
   Move(Next,P^,SizeOf(Next)); Move(Rec,P^[SizeOf(Next)],Size);
   Next:=P; Add:=True;
  end;
 end;
function TVirtualList.Del(var Rec):boolean;
 var P:^pointer; D:PBlock; OK:boolean;
 begin
  P:=addr(Next); OK:=False;
  while (not OK) and (P^<>nil) do begin
   D:=P^; OK:=TestEl(Rec,D^[SizeOf(Next)]);
   if OK then begin
    Move(D^,P^,SizeOf(Next));
    FreeMem(D,Size+SizeOf(Next));
   end else P:=P^;
  end;
  Del:=OK;
 end;

constructor TRecProtList.Init; begin inherited Init(SizeOf(longint)) end;
function TRecProtList.TestEl(var Rec,Patt):boolean;
 var
  I:longint absolute Rec;
  J:longint absolute Patt;
 begin TestEl:=I=J end;

{------------------------------------------------------}

constructor TDataFile.Open(var DFNR);
 var i:byte; Fld:TFldRec; fi:TRecFileFastIn; PB:PBlock;
 begin
  DFNameRec(DFNR).Put(@Self); RecSize:=1; InitFieldsFile;
{ WrLogStr('Fields^.Length=$'+HexLong(Fields^.Length)); }
  with Fields^ do if Length>0 then for i:=0 to Length-1 do begin
   GetRec(i,Fld); inc(RecSize,Fld.Size);
{
   WrLogStr('SizeOfTFldRec=$'+HexLong(SizeOf(TFldRec))+' '+
            'SizeOfFld=$'+HexLong(SizeOf(Fld))+' '+
            'FldNo=$'+HexByte(i)+' '+
            'Fld.Size=$'+HexByte(Fld.Size));
}
  end;
  { Eval actual file length }
  GetMem(PB,RecSize); fi.Init(FName+'.dat',RecSize); OpenLen:=0;
  while not fi.e_file do
   begin fi.GetNext(PB^); if PB^[0]<>$FF then inc(OpenLen) end;
  fi.Done; FreeMem(PB,RecSize);
  InitDataFile;
  for i:=0 to FldMaxCnt do IndArr[i]:=Nil; OpenIndex;
  LastFileLink:=0; RecProtList.Init;
 end;
destructor TDataFile.Kill;
 var i:byte;
 begin
  for i:=0 to FldMaxCnt do if IndArr[i]<>Nil then
   begin Dispose(IndArr[i],Done); FileDelete(IndexName(i)) end;
  Dispose(Data,  Done); FileDelete(FName+'.dat');
  Dispose(Fields,Done); FileDelete(FName+'.fld');
 end;
constructor TDataFile.Create(var DFNR);
 begin
  with DFNameRec(DFNR) do
   begin FileDelete(FN+'.dat'); FileDelete(FN+'.fld') end;
  Open(DFNR);
 end;
procedure TDataFile.InitDataFile;
 begin New(Data,Init(FName+'.dat',RecSize,StdDataBufLen)) end;
procedure TDataFile.InitFieldsFile;
 begin New(Fields,Init(FName+'.fld',SizeOf(TFldRec),FldMaxCnt+1)) end;
procedure TDataFile.MakeCopy(FN:SmallFName);
 procedure FileCopy(PF:string; aSize:word);
  var fi:TRecFileFastIn; fo:TRecFileFastOut; PB:PBlock;
  begin
   fi.Init(FName+PF,aSize); fo.Init(FN+PF,aSize); GetMem(PB,aSize);
   while not fi.e_file do begin fi.GetNext(PB^); fo.PutNext(PB^) end;
   fi.Done; fo.Done; FreeMem(PB,aSize);
  end;
 begin
  Dispose(Data,Done); FileCopy('.dat',RecSize); InitDataFile;
  Dispose(Fields,Done); FileCopy('.fld',SizeOf(TFldRec)); InitFieldsFile;
 end;
function TDataFile.LessP(p:longint; FldId:byte; Dir:boolean):boolean;
 var W:PBlock; Fld:TFldRec;
 begin
  GetMem(W,RecSize);
  Data^.GetRec(p,W^); Fields^.GetRec(FldId,Fld);
  if Dir
   then LessP:=FldLess(Fld,PLessP^,W^[CLessP])
   else LessP:=FldLess(Fld,W^[CLessP],PLessP^);
  FreeMem(W,RecSize);
 end;
function TDataFile.FindFirst(FldId:byte; var R; Dir:boolean):longint;
 begin
  PLessP:=addr(R); CLessP:=FldPos(FldId);
  FindFirst:=IndArr[FldId]^.FindFirst(Dir);
 end;
function TDataFile.IndexName(FldId:byte):string;
 var s:string[2]; FN:String[20];
 begin Str(FldId:2,s); if s[1]=' ' then s[1]:='0'; IndexName:=FName+'.i'+s end;
procedure TDataFile.EraseIndex(FldId:byte);
 begin 
  WrLogStr('EraInd start ...');
  Dispose(IndArr[FldId],Kill); 
  WrLogStr('EraInd Kill OK ...');
  IndArr[FldId]:=Nil 
 end;
procedure TDataFile.CreateIndex(FldId:byte);
 begin
  Application^.Insert(New(PIndexingWindow,Init(@Self,FldId)));
  Message(Application,evBroadcast,cmCreateIndex,Nil);
 end;
procedure TDataFile.OpenIndex;
 var i:byte; Fld:TFldRec;
 begin if Fields^.Length>0 then for i:=0 to Fields^.Length-1 do begin
  Fields^.GetRec(i,Fld);
  if FldGetFlags(Fld,ffIndexed) then begin
   New(IndArr[i],Init(IndexName(i),addr(Self),i));
   if OpenLen<>IndArr[i]^.Data.Cnt then begin
{
    WrLogStr('OpenLen=$'+HexLong(OpenLen)+' '+
             'Index[$'+HexByte(i)+']=$'+HexLong(IndArr[i]^.Data.Cnt));
}
    Dispose(IndArr[i],Done); CreateIndex(i)
   end;
  end;
 end end;
procedure TDataFile.SetFldLinks;
 var i:byte; Fld:TFldRec;
 begin if Fields^.Length>0 then begin
  for i:=0 to Fields^.Length-1 do begin
   Fields^.GetRec(i,Fld);
   with Fld do if FType=CFPntr then begin
    FPtr:=DF_FindId(FPName);
    if FPtr=$ff then
     begin LifeError('File '+FPName+' is not defined'); Halt end;
   end;
   Fields^.PutRec0(i,Fld);
  end;
  Fields^.Flush;
 end end;
{$IFDEF Turbo} {$F+} {$ENDIF}
var SFld:TFldRec;
function Fld_Less(var o1,o2):boolean;
 const pos = SizeOf(longint);
 var p1:longint absolute o1; p2:longint absolute o2;
     b1:block   absolute o1; b2:block   absolute o2;
 begin
  if FldEqual(SFld,b1[pos],b2[pos])
   then Fld_Less:=p1<=p2 else Fld_Less:=FldLess(SFld,b1[pos],b2[pos]);
 end;
{$IFDEF Turbo} {$F-} {$ENDIF}
procedure TDataFile.ExtrIndData(
 FldId:byte; TmpIName:string; var TmpISize:word; PV:PView);
 var
  PB,PI:PBlock; c:word; N:longint;
  fi:TRecFileFastIn; fo:TRecFileFastOut;
 begin
  Fields^.GetRec(FldId,SFld); c:=FldPos(FldId);
  Dispose(Data,Done); GetMem(PB,RecSize);
  TmpISize:=SizeOf(longint)+SFld.Size; N:=0;
  GetMem(PI,TmpISize);
  fi.Init(FName+'.dat',RecSize);
  fo.Init(TmpIName,TmpISize);
  while not fi.e_file do begin
   fi.GetNext(PB^);
   if PB^[0]<>$FF then begin
    Move(N,PI^[0],SizeOf(Longint));
    Move(PB^[c],PI^[SizeOf(longint)],SFld.Size);
    fo.PutNext(PI^)
   end;
   Inc(N);
  end;
  fo.Done; fi.Done;
  FreeMem(PI,TmpISize); FreeMem(PB,RecSize); InitDataFile;
  Sort_File(TmpIName,TmpISize,Fld_Less,PV);
 end;
procedure TDataFile.AppendField(FldId:byte; var Fld:TFldRec);
 var
  PB:PBlock; L:longint;
  fi:TRecFileFastIn; fo:TRecFileFastOut; f:file;
 begin
  GetMem(PB,RecSize+Fld.Size); FldStdClear(Fld,PB^[RecSize]); L:=Data^.Length;
  if L>0 then begin
   SetUseMode; EditFirstFldValue(Fld,PB^[RecSize],0); SetDesignMode end;
  { Changes in structure starts here: }
  Fields^.PutRec(FldId,Fld);
  Dispose(Data,Done);
  if L>0 then begin
   fi.Init(FName+'.dat',RecSize);
   fo.Init(FName+'.new',RecSize+Fld.Size);
   while not fi.e_file do
    begin fi.GetNext(PB^); fo.PutNext(PB^) end;
   fo.Done; fi.Kill; Assign(f,FName+'.new'); Rename(f,FName+'.dat');
  end;
  Inc(RecSize,Fld.Size); InitDataFile;
  FreeMem(PB,RecSize);
  if FldGetFlags(Fld,ffIndexed) then CreateIndex(FldId);
 end;
procedure TDataFile.EraseField(FldNum:byte);
 var
  Fld:TFldRec; c:word; L:longint; PB:PBlock;
  fi:TRecFileFastIn; fo:TRecFileFastOut; f:file;
 procedure UpdateData;
  begin
   c:=FldPos(FldNum);
   L:=Data^.Length; Dispose(Data,Done);
   if L>0 then begin
    GetMem(PB,RecSize);
    fi.Init(FName+'.dat',RecSize);
    fo.Init(FName+'.new',RecSize-Fld.Size);
    while not fi.e_file do begin
     fi.GetNext(PB^);
     Move(PB^[c+Fld.Size],PB^[c],RecSize-c-Fld.Size);
     fo.PutNext(PB^)
    end;
    fo.Done; fi.Kill;
    Assign(f,FName+'.new'); Rename(f,FName+'.dat');
    FreeMem(PB,RecSize);
   end;
   Dec(RecSize,Fld.Size); InitDataFile;
  end;
 procedure UpdateFields;
  var i:byte;
  begin with Fields^ do begin
   if FldNum+1<Length then for i:=FldNum+1 to Length-1 do begin
    GetRec(i,Fld); PutRec0(i-1,Fld);
    IndArr[i-1]:=IndArr[i]; IndArr[i]:=Nil;
    if FldGetFlags(Fld,ffIndexed) then IndArr[i-1]^.ChangeName(IndexName(i-1),i-1);
   end;
   Trunc0(Length-1); Flush;
  end end;
 begin
  Fields^.GetRec(FldNum,Fld);
  if FldGetFlags(Fld,ffIndexed) then EraseIndex(FldNum);
  UpdateData; UpdateFields;
 end;
procedure TDataFile.SwapField(FldNum:byte);
 var
  Fld0,Fld1:TFldRec; c:word; L:longint; PB:PBlock; W:Small;
  fi:TRecFileFastIn; fo:TRecFileFastOut; f:file;
 procedure UpdateData;
  begin
   c:=FldPos(FldNum); L:=Data^.Length; Dispose(Data,Done);
   if L>0 then begin
    GetMem(PB,RecSize);
    fi.Init(FName+'.dat',RecSize);
    fo.Init(FName+'.new',RecSize);
    while not fi.e_file do begin
     fi.GetNext(PB^);
     Move(PB^[c],W,Fld0.Size);
     Move(PB^[c+Fld0.Size],PB^[c],Fld1.Size);
     Move(W,PB^[c+Fld1.Size],Fld0.Size);
     fo.PutNext(PB^)
    end;
    fo.Done; fi.Kill;
    Assign(f,FName+'.new'); Rename(f,FName+'.dat');
    FreeMem(PB,RecSize);
   end;
   InitDataFile;
  end;
 var PI:PIndex;
 begin
  Fields^.GetRec(FldNum,Fld0); Fields^.GetRec(FldNum+1,Fld1);
  { Swap data }
  UpdateData;
  { Swap Fields }
  Fields^.PutRec(FldNum,Fld1); Fields^.PutRec(FldNum+1,Fld0);
  { Swap Indexs }
  PI:=IndArr[FldNum]; if FldGetFlags(Fld0,ffIndexed)
   then PI^.ChangeName('Index.Tmp',99);
  IndArr[FldNum]:=IndArr[FldNum+1];
  if FldGetFlags(Fld1,ffIndexed)
   then IndArr[FldNum]^.ChangeName(IndexName(FldNum),FldNum);
  if FldGetFlags(Fld0,ffIndexed)
   then PI^.ChangeName(IndexName(FldNum+1),FldNum+1);
  IndArr[FldNum+1]:=PI;
 end;
procedure TDataFile.ChangeField(FldId:byte; var OFld,NFld:TFldRec);
 var OK:boolean; L:longint;
 procedure Convert;
  var
   c0,c1,c2,Newrs,delta:word; PB0,PB1:PBlock;
   fi:TRecFileFastIn; fo:TRecFileFastOut; f:file;
  procedure ConvField(var A,B);
   var
    b0:byte absolute A;    b1:byte absolute B;
    w0:word absolute A;    w1:word absolute B;
    l0:longint absolute A; l1:longint absolute B;
    r0:real absolute A;    r1:real absolute B;
   procedure UpFirst;
    var i:byte; S:Block absolute B; AS0:boolean;
    begin
     AS0:=True;
     if NFld.Size>0 then for i:=0 to NFld.Size-1 do begin
      if AS0 and (S[i] in [ord('a')..ord('z'),ord('')..ord('')])
       then dec(S[i],32);
      AS0:=S[i]=ord(' ');
     end;
    end;
   begin case OFld.FType of
    CFList:case NFld.FType of
            CFWord:w1:=word(b0); CFLong:l1:=longint(b0); CFReal:r1:=b0;
           end;
    CFWord:case NFld.FType of
            CFList:b1:=byte(w0); CFLong:l1:=longint(w0); CFReal:r1:=w0;
           end;
    CFLong:case NFld.FType of
            CFList:b1:=byte(l0); CFWord:w1:=word(l0); CFReal:r1:=l0;
           end;
    CFReal:case NFld.FType of
            CFList:b1:=byte(round(r0)); CFWord:w1:=word(round(r0)); CFLong:l1:=round(r0);
           end;
    CFStr,CFName:begin
            FillChar(B,NFld.Size,' ');
            if OFld.Size<NFld.Size
             then Move(A,B,OFld.Size) else Move(A,B,NFld.Size);
            if NFld.FType=CFName then UpFirst;
           end;
   end end;
  begin
    WrLogStr('ConvertFld start ...');
    c0:=FldPos(FldId); c1:=c0+OFld.Size; c2:=c0+NFld.Size;
    NewRS:=RecSize+NFld.Size-OFld.Size; delta:=RecSize-c1;
    WrLogStr('ConvertFld eval positions OK ...');
    GetMem(PB0,RecSize); GetMem(PB1,NewRS);
    WrLogStr('ConvertFld getmem OK ...');
    fi.Init(FName+'.dat',RecSize);
    fo.Init(FName+'.new',NewRS);
    WrLogStr('ConvertFld open files OK ...');
    while not fi.e_file do begin
     fi.GetNext(PB0^);
     Move(PB0^[0],PB1^[0],c0);
     if delta>0 then Move(PB0^[c1],PB1^[c2],delta);
     ConvField(PB0^[c0],PB1^[c0]);
     fo.PutNext(PB1^)
    end;
    WrLogStr('ConvertFld convert files OK ...');
    fo.Done; fi.Kill;
    WrLogStr('ConvertFld close files OK ...');
    Assign(f,FName+'.new'); Rename(f,FName+'.dat');
    WrLogStr('ConvertFld rename files OK ...');
    FreeMem(PB0,RecSize); FreeMem(PB1,NewRS);
    WrLogStr('ConvertFld OK ...');
  end;
 begin
  WrLogStr('ChangeFld start ...');
  if OFld.Size>NFld.Size then begin
   ErrNum(esNewFldShort); OK:=YN_Num2(esMayLousData,esDoTransform);
  end else OK:=True;
  if OK then begin
   WrLogStr('ChangeFld erasing index ...');
   if FldGetFlags(OFld,ffIndexed) then EraseIndex(FldId);
   WrLogStr('ChangeFld disposing Data ...');
   L:=Data^.Length; Dispose(Data,Done);
   WrLogStr('ChangeFld convertion begin ...');
   if L>0 then Convert;
   WrLogStr('ChangeFld convertion end ...');
   Dec(RecSize,OFld.Size); Inc(RecSize,NFld.Size);
   WrLogStr('ChangeFld change RecSize OK ...');
   InitDataFile;
   WrLogStr('ChangeFld InitDataFile OK ...');
   Fields^.PutRec(FldId,NFld);
   WrLogStr('ChangeFld Save New field OK ...');
   if FldGetFlags(NFld,ffIndexed) then CreateIndex(FldId);
   WrLogStr('ChangeFld create index OK ...');
  end;
 end;
function TDataFile.FldName(FldId:byte):string;
 var Fld:TFldRec;
 begin
  Fields^.GetRec(FldId,Fld);
  FldName:=ClearSp(RecStr(Fld.Name,FldMaxLen));
 end;
function TDataFile.FldNameLnks(FldId:byte):string;
 var Fld:TFldRec;
 begin
  Fields^.GetRec(FldId,Fld);
  FldNameLnks:=
{$IFDEF Cyr}
   ' '+ClearSp(RecStr(Fld.CntName,6))+
   ' ( '+LName+'.'+FldName(FldId)+')';
{$ENDIF}
{$IFDEF Lat}
   'linked '+ClearSp(RecStr(Fld.CntName,6))+
   ' (by '+LName+'.'+FldName(FldId)+')';
{$ENDIF}
 end;
function TDataFile.LinkChois:byte;
 var P:PChoisList; i:byte;
 begin
  New(P,Init);
  for i:=0 to FLCnt-1 do with FLArr[i]
   do P^.Insert(DFArr[FileId]^.FldNameLnks(FldId));
  P^.SetCursor(LastFileLink);
  {$IFDEF Turbo} Dispose(P,Chois(i)); {$ENDIF}
  {$IFDEF FPC} P^.Chois(i); {$ENDIF}
  LinkChois:=i;
 end;
function TDataFile.PntrChois:byte;
 var P:PChoisList; i,j,k:byte;
 begin
  New(P,Init);
  for i:=0 to Fields^.Length-1 do if FldType(i)=CFPntr
{$IFDEF Cyr}
   then P^.Insert('  '+FldName(i));
{$ENDIF}
{$IFDEF Lat}
   then P^.Insert('projected on '+FldName(i));
{$ENDIF}
  {$IFDEF Turbo} Dispose(P,Chois(j)); {$ENDIF}
  {$IFDEF FPC} P^.Chois(j); {$ENDIF}
  PntrChois:=$ff; k:=0;
  for i:=0 to Fields^.Length-1 do if FldType(i)=CFPntr
   then begin if k=j then PntrChois:=i; inc(k) end;
 end;
function TDataFile.FldType(FldId:byte):byte;
 var Fld:TFldRec;
 begin
  Fields^.GetRec(FldId,Fld);
  FldType:=Fld.FType;
 end;
function TDataFile.FldULin(FldId:byte):boolean;
 var Fld:TFldRec;
 begin
  Fields^.GetRec(FldId,Fld);
  FldULin:=FldGetFlags(Fld,ffFulShow);
 end;
procedure TDataFile.SelectIndex(var IndexId:byte);
 var P:PChoisList; i,k:byte; A:Small; Fld:TFldRec;
 begin
  New(P,Init);
  k:=0;
  for i:=0 to Fields^.Length-1 do begin
   Fields^.GetRec(i,Fld);
   if FldGetFlags(Fld,ffIndexed) then begin
    A[k]:=i;
{$IFDEF Cyr}
    P^.Insert('  '+FldName(i));
{$ENDIF}
{$IFDEF Lat}
    P^.Insert('sorted by '+FldName(i));
{$ENDIF}
    if i=IndexId then P^.SetCursor(k); Inc(k);
   end;
  end;
  A[k]:=$FF; P^.Insert(ErrStr(esUnSortOrder));
  {$IFDEF Turbo} Dispose(P,Chois(i)); {$ENDIF}
  {$IFDEF FPC} P^.Chois(i); {$ENDIF}
  if i<=k then IndexId:=A[i];
 end;
function TDataFile.SelectFIndex:byte;
 var i,k:byte; Fld:TFldRec;
 begin
  k:=LastUsedIndex; i:=0;
  while (k=$FF) and (i<Fields^.Length) do begin
   Fields^.GetRec(i,Fld);
   if FldGetFlags(Fld,ffRootInd) then k:=i;
   Inc(i);
  end;
  SelectFIndex:=k;
 end;
function TDataFile.StdSummLine(ExFld:byte):string;
 var ps:PString; i,l,k:byte; Fld:TFldRec;
 begin
  New(ps); if CntSize=0 then ps^:='' else ps^:=Blank(CntSize)+VLChar;
  for i:=0 to Fields^.Length-1 do if i<>ExFld then begin
   Fields^.GetRec(i,Fld); l:=Fld.ShowSize;
   if l>0 then if FldGetFlags(Fld,ffCntSumm)
    then begin for k:=1 to l do ps^:=ps^+'?'; ps^:=ps^+VLChar end
    else ps^:=ps^+Blank(l)+VLChar;
  end;
  if FLCnt>0 then for i:=0 to FLCnt-1 do ps^:=ps^+Blank(6)+VLChar;
  StdSummLine:=ps^; Dispose(ps);
 end;
function TDataFile.StdHeaderLine(ExFld:byte; var SummLineExist:boolean):string;
 procedure AddPSummColumn(var s:string; FId:byte);
  var i:byte; Fld:TFldRec;
  begin with DFArr[FId]^.Fields^ do for i:=0 to Length-1 do begin
   GetRec(i,Fld);
   if FldGetFlags(Fld,ffCntPntr)
    then s:=s+JustStr(' '+RecStr(Fld.Name,FldMaxLen),Fld.ShowSize)+VLChar;
  end end;
 var ps,ps1:PString; i,l:byte; Fld:TFldRec; ns:string[9];
 begin
  New(ps); New(ps1);
  if CntSize=0
   then ps^:=''
   else begin
    ns:='No'; if CntSize>3 then ns:=ns+' ';
    LExpandStr(ns,CntSize,' '); ps^:=ns+VLChar;
   end;
  SummLineExist:=false;
  for i:=0 to Fields^.Length-1 do if i<>ExFld then begin
   Fields^.GetRec(i,Fld); l:=Fld.ShowSize;
   if l>0 then begin
    if FldGetFlags(Fld,ffCntSumm) then SummLineExist:=True;
    ps1^:=FldName(i);
    if length(ps1^)>l
     then Delete(ps1^,l+1,length(ps1^)-l)
     else while length(ps1^)<l do
      begin ps1^:=' '+ps1^; if length(ps1^)<l then ps1^:=ps1^+' ' end;
    ps^:=ps^+ps1^+VLChar;
   end;
  end;
  if FLCnt>0 then for i:=0 to FLCnt-1 do begin
   DFArr[FLArr[i].FileId]^.Fields^.GetRec(FLArr[i].FldId,Fld);
   ps^:=ps^+RecStr(Fld.CntName,6)+VLChar;
   AddPSummColumn(ps^,FLArr[i].FileId);
  end;
  StdHeaderLine:=ps^;
  Dispose(ps); Dispose(ps1);
 end;
function TDataFile.GetCount(Ind:longint; FileId,FldId:byte):longint;
 var N0:longint;
 begin
  N0:=Ind;
  GetCount:=DFArr[FileId]^.FindFirst(FldId,N0,False)+1-
            DFArr[FileId]^.FindFirst(FldId,N0,True);
 end;
function TDataFile.GetPSumm(Ind:longint; FldId:byte):string;
 var N0,NS,NE,i:longint; PB,PR:PBlock; MaxFld,j:byte;
  FPos:array[0..FldMaxCnt] of word;
  FTyp:array[0..FldMaxCnt] of byte;
  FFlg:array[0..FldMaxCnt] of boolean;
 procedure Prepare;
  var c:word; j:byte; Fld:TFldRec;
  begin
   MaxFld:=Fields^.Length-1; c:=1;
   for j:=0 to MaxFld do begin
    Fields^.GetRec(j,Fld);
    FPos[j]:=c; FTyp[j]:=Fld.FType;
    FFlg[j]:=FldGetFlags(Fld,ffCntPntr);
    inc(c,Fld.Size);
   end;
  end;
 procedure AddNum(t:byte; var s,d);
  var rs:real absolute s; rd:real absolute d;
      ls:longint absolute s; ld:longint absolute d;
  begin case t of CFReal:rd:=rd+rs; CFLong:ld:=ld+ls; end end;
 function StrPSumm:string;
  var s:string; j:byte; Fld:TFldRec;
  begin
   s:='';
   for j:=0 to MaxFld do if FFlg[j] then begin
    Fields^.GetRec(j,Fld);
    s:=s+FldStdBrouseStr(Fld,PB^[FPos[j]])+VLChar;
   end;
   StrPSumm:=s
  end;
 begin if PntrSummExists then begin
  N0:=Ind; NS:=FindFirst(FldId,N0,True); NE:=FindFirst(FldId,N0,False);
  GetMem(PB,RecSize); GetMem(PR,RecSize);
  RecClear(PB^); Prepare;
  if NE>=NS then for i:=NS to NE do begin
   Data^.GetRec(IndArr[FldId]^.GetV(i),PR^);
   for j:=0 to MaxFld do if FFlg[j]
    then AddNum(FTyp[j],PR^[FPos[j]],PB^[FPos[j]]);
  end;
  GetPSumm:=StrPSumm; 
  FreeMem(PR,RecSize); FreeMem(PB,RecSize);
 end else GetPSumm:='' end;
function TDataFile.StdBrouseLine(
 Ind,Num:longint; var Rec; ExFld:byte; SummLine:boolean):string;
 var
  ps:PString; s1:string[6]; i:byte; c:word; Fld:TFldRec;
  B:Block absolute Rec; ns:string[10];
 begin
  New(ps); c:=1;
  if CntSize=0 then ps^:='' else begin
   if Num>=0 then Str(Num+1,ns) else ns:='';
   while Length(ns)>CntSize do Delete(ns,1,1);
   LExpandStr(ns,CntSize,' '); ps^:=ns+VLChar
  end;
  for i:=0 to Fields^.Length-1 do begin
   Fields^.GetRec(i,Fld);
   if (i<>ExFld) and (Fld.ShowSize>0) then begin
    if ((not SummLine) and (Ind<>-1)) or
       (SummLine and FldGetFlags(Fld,ffCntSumm))
     then ps^:=ps^+FldStdBrouseStr(Fld,B[c])
     else ps^:=ps^+Blank(Fld.ShowSize);
    ps^:=ps^+VLChar;
   end;
   inc(c,Fld.Size);
  end;
  if FLCnt>0 then for i:=0 to FLCnt-1 do with FLArr[i] do if SummLine
   then ps^:=ps^+Blank(6)+VLChar
   else begin
    Str(GetCount(Ind,FileId,FldId):6,s1); ps^:=ps^+s1+VLChar;
    ps^:=ps^+DFArr[FileId]^.GetPSumm(Ind,FldId);
   end;
  StdBrouseLine:=ps^; Dispose(ps);
 end;
function TDataFile.StdPtrLine(p:longint; Depth:byte):string;
 var
  ps:PString; i:byte; c:word; Fld:TFldRec;
  PB:PBlock; brackets:boolean;
 begin if p=-1 then StdPtrLine:='---' else begin
  New(ps); GetMem(PB,RecSize); Data^.GetRec(p,PB^);
  ps^:=''; c:=1; i:=0;
  if Depth>0 then while (i<Fields^.Length) and (length(ps^)<100) do begin
   Fields^.GetRec(i,Fld); if i>0 then ps^:=ps^+'; ';
   brackets:=Fld.Ftype=CFPntr;
   if brackets then brackets:=DFArr[Fld.FPtr]^.Fields^.Length>1;
   if brackets then ps^:=ps^+'[';
   ps^:=ps^+ClearSp(FldStdString(Fld,PB^[c],Depth-1));
   if brackets then ps^:=ps^+']';
   inc(c,Fld.Size); Inc(i);
  end;
  StdPtrLine:=ps^; FreeMem(PB,RecSize);
  Dispose(ps);
 end end;
procedure TDataFile.RecClear(var Rec);
 var i:byte; c:word; Fld:TFldRec; B:Block absolute Rec;
 begin
  c:=1; B[0]:=0;
  for i:=0 to Fields^.Length-1 do begin
   Fields^.GetRec(i,Fld); FldStdClear(Fld,B[c]); inc(c,Fld.Size);
  end;
 end;
procedure TDataFile.TestForUnique(ExI:Longint; var NewR; var SucInd:longint);
 type
  TUInt = record FId:byte; S,E:longint end;
 var
  A:array[0..FldMaxCnt] of TUInt; Cnt,FMin:byte;
  Fld:TFldRec; c:word; i:byte; NB:Block absolute NewR;
  L,N,Place:longint; NTst,OK:boolean;
 begin
  { Eval Unique intervals }
  c:=1; Cnt:=0; L:=999999; FMin:=$FF;
  for i:=0 to Fields^.Length-1 do begin
   Fields^.GetRec(i,Fld);
   if FldGetFlags(Fld,ffUniTest) then with A[Cnt] do begin
    S:=FindFirst(i,NB[c],True); E:=FindFirst(i,NB[c],False)+1;
    if L>E-S then begin L:=E-S; FMin:=Cnt end;
    FId:=i; Inc(Cnt);
   end;
   inc(c,Fld.Size);
  end;
  { Test smallest Unique interval }
  if FMin<>$FF then if A[FMin].E>A[FMin].S then begin
   L:=A[FMin].S;
   while (L<A[FMin].E) and (SucInd<0) do begin
    N:=IndArr[A[FMin].FId]^.GetV(L);
    if N<>ExI then begin
     NTst:=True;
     if Cnt>1 then for i:=0 to Cnt-1 do if (i<>FMin) and NTst then begin
      Place:=IndArr[A[i].FId]^.Find(N,OK);
      if not OK then ErrNum(esInpossible0)
       else if (Place<A[i].S) or (Place>=A[i].E) then NTst:=False;
     end;
     if NTst then SucInd:=N;
    end;
    Inc(L);
   end;
  end;
 end;
procedure TDataFile.RecUpdate(var Ind:longint; var NewR; old,always:boolean);
 procedure UpdateGo;
  var
   i:byte; c:word; Fld:TFldRec;
   POR:PBlock; NB:Block absolute NewR;
  begin
   GetMem(POR,RecSize);
   { if old rec exist, clear indexes for changed fields }
   if old then begin
    c:=1; Data^.GetRec(Ind,POR^);
    for i:=0 to Fields^.Length-1 do begin
     Fields^.GetRec(i,Fld);
     if FldGetFlags(Fld,ffIndexed) and (not FldEqual(Fld,POR^[c],NB[c]))
      then IndArr[i]^.Del(Ind);
     inc(c,Fld.Size);
    end;
    Data^.PutRec(Ind,NewR);
   end else Ind:=Data^.Append(NewR);
   c:=1; { add indexes for changed fields (for all if new rec) }
   for i:=0 to Fields^.Length-1 do begin
    Fields^.GetRec(i,Fld);
    if FldGetFlags(Fld,ffIndexed) and
       ((not old) or (not FldEqual(Fld,POR^[c],NB[c])))
     then IndArr[i]^.Add(Ind);
    inc(c,Fld.Size);
   end;
   FreeMem(POR,RecSize);
  end;
 var SucInd:longint; s:string[9];
 begin
  if not old then Ind:=-1; SucInd:=-1;
  TestForUnique(Ind,NewR,SucInd);
  if SucInd<0
   then UpdateGo
   else begin
    if Not Always then begin
     Str(SucInd,s); LifeError(ErrStr(esBadDupAttmp)+LName+'['+s+']') end;
    Ind:=SucInd;
   end;
 end;
function TDataFile.ExchRelated(Ind,Dest:longint):boolean;
 var i:byte; N0,N1,N2:longint; c:word; PB:PBlock; OK:boolean;
 begin
  OK:=True;
  if FLCnt>0 then for i:=0 to FLCnt-1 do with FLArr[i] do begin
   GetMem(PB,DFArr[FileId]^.RecSize);
   c:=DFArr[FileId]^.FldPos(FldId);
   while (GetCount(Ind,FileId,FldId)>0) and OK do begin
    N0:=Ind; N1:=DFArr[FileId]^.FindFirst(FldId,N0,True);
    N2:=DFArr[FileId]^.IndArr[FldId]^.GetV(N1);
    if not DFArr[FileId]^.RecProtList.Test(N2) then begin
     DFArr[FileId]^.Data^.GetRec(N2,PB^);
     DFArr[FileId]^.IndArr[FldId]^.Del(N2);
     Move(Dest,PB^[c],SizeOf(longint));
     DFArr[FileId]^.Data^.PutRec(N2,PB^);
     DFArr[FileId]^.IndArr[FldId]^.Add(N2);
    end else OK:=False;
   end;
   FreeMem(PB,DFArr[FileId]^.RecSize);
  end;
  ExchRelated:=OK;
 end;
function TDataFile.RecErase(Ind,Dest:longint):boolean;
 var  PB:PBlock; i:byte;
 begin if (Ind<>Dest) and (not RecProtList.Test(Ind))
  then if ExchRelated(Ind,Dest)
   then begin
    for i:=0 to Fields^.Length-1 do
     if IndArr[i]<>Nil then IndArr[i]^.Del(Ind);
    GetMem(PB,RecSize);
    Data^.GetRec(Ind,PB^); PB^[0]:=$FF; Data^.PutRec(Ind,PB^);
    FreeMem(PB,RecSize);
    RecErase:=True;
   end
   else begin ErrNum(esLnkRecInUse); RecErase:=False end
  else begin ErrNum(esRecIsNowUsd); RecErase:=False end;
 end;
function TDataFile.RelationsCount:byte;
 var i,c:byte; Fld:TFldRec;
 begin
  c:=0;
  if Fields^.Length>0 then for i:=0 to Fields^.Length-1 do begin
   Fields^.GetRec(i,Fld);
   if Fld.FType=CFPntr then inc(c);
  end;
  RelationsCount:=c;
 end;
function TDataFile.FullShowCount:byte;
 var i,c:byte; Fld:TFldRec;
 begin
  c:=0;
  if Fields^.Length>0 then for i:=0 to Fields^.Length-1 do
   begin Fields^.GetRec(i,Fld); if FldGetFlags(Fld,ffFulShow) then inc(c) end;
  FullShowCount:=c;
 end;
function TDataFile.FldPos(FldId:byte):word;
 var i,c:word; Fld:TFldRec;
 begin
  i:=0; c:=1;
  while i<FldId do begin Fields^.GetRec(i,Fld); inc(c,Fld.Size); inc(i) end;
  FldPos:=c;
 end;
function TDataFile.Less(p1,p2:longint; FldId:byte):boolean;
 var c:word; Fld:TFldRec; b1,b2:PBlock;
 begin
  Fields^.GetRec(FldId,Fld);
  if FldGetFlags(Fld,ffIndexed) then begin
   c:=FldPos(FldId);
   GetMem(b1,RecSize); GetMem(b2,RecSize);
   Data^.GetRec(p1,b1^); Data^.GetRec(p2,b2^);
   if FldEqual(Fld,b1^[c],b2^[c])
    then Less:=p1<=p2
    else Less:=FldLess(Fld,b1^[c],b2^[c]);
   FreeMem(b1,RecSize); FreeMem(b2,RecSize);
  end else Less:=p1<=p2;
 end;

function DF_FindId(FN:SmallFName):byte;
 var i,j:byte;
 begin
  j:=$FF;
  for i:=254 downto 0 do if DFArr[i]=Nil
   then else if DFArr[i]^.FName=FN then j:=i;
  DF_FindId:=j;
 end;
function GetFileId(FLCursor:longint):byte;
 var i,j:byte;
 begin
  i:=0; j:=0;
  if (FLCursor>=0) and (FLCursor<FileCount)
   then repeat
    while DFArr[i]=Nil do Inc(i);
    if j=FLCursor then GetFileId:=i;
    inc(j); inc(i);
   until j>FLCursor
   else GetFileId:=$FF;
 end;

procedure SetUseMode;
 var i,j:byte; Fld:TFldRec;
 begin if SystemMode<>UseMode then begin
  SystemMode:=UseMode;
  ClearFileLinks;
  for i:=0 to 254 do if DFArr[i]<>Nil then with DFArr[i]^ do begin
   PntrSummExists:=False;
   with Fields^ do if Length>0 then
    for j:=0 to Length-1 do begin
     GetRec(j,Fld);
     if Fld.FType=CFPntr then AddFileLink(i,Fld.FPtr,j);
     if FldGetFlags(Fld,ffCntPntr) then PntrSummExists:=True;
    end;
   LastUsedIndex:=$FF;
  end;
 end end;
procedure SetDesignMode;
 begin SystemMode:=DesignMode end;

procedure ClearFileLinks;
 var i:byte;
 begin for i:=0 to 254 do if DFArr[i]<>Nil then DFArr[i]^.FLCnt:=0 end;
procedure AddFileLink(FlSors,FLDest,FieldId:byte);
 begin with DFArr[FlDest]^ do if FLCnt>FileLinkMax
  then begin LifeError(ErrStr(esManyLinksTo)+LName); Halt end
  else begin
   with FLArr[FLCnt] do begin FileId:=FlSors; FldId:=FieldId end;
   Inc(FLCnt);
  end;
 end;

{------------------ Index drivers: -----------------------}

const
 GMax = 20;
var
 GCnt:byte;
 GArr:array[0..GMax] of longint;

procedure TIPartClear(var TIP:TIPart);
 begin with TIP do begin Cnt:=0; Ptr:=0 end end;

procedure TINodeClear(var TIN:TINode; aV:TValue; Root:longint);
 begin with TIN do begin
  V:=aV; TIPartClear(L); TIPartClear(R); B:=Root
 end end;

constructor TIndex.Init(aName:FullFName; aPDF:PDataFile; aFldId:byte);
 var A:TINode;
 begin
  PDF:=aPDF; FldId:=aFldId;
  inherited Init(aName,SizeOf(TINode),StdIndxBufLen);
  TIPartClear(Data); N:=Length;
  case N of
   0:begin TINodeClear(A,0,0); PutRec(0,A) end;
   1:begin
      GetRec(0,A); Data.Cnt:=A.B;
      if Data.Cnt=1 then Data.Ptr:=A.V else Data.Ptr:=0;
      if Data.Cnt<2 then dec(N);
     end;
  end;
  if N>0 then begin GetRec(0,A); Data.Ptr:=0; Data.Cnt:=A.B end;
 end;
constructor TIndex.Build(
 aName:FullFName; aPDF:PDataFile; aFldId:byte;
 TmpIName:FullFName; TmpISize:word; PV:PView);
 var F:TRecFileDrv; Depth:byte; PB:PBlock;
 procedure BuildPart(var P:TIPart; Rt:longint; S,E:longint);
  var A:TINode; M,V:longint;
  begin with P do begin
   Cnt:=E-S;
   if Cnt=0 then Ptr:=0 else begin
    M:=(S+E) shr 1; F.GetRec(M,PB^); Move(PB^,V,SizeOf(longint));
    if Cnt=1 then Ptr:=V else begin
     PV^.WriteChar(17+Depth,4,'>',1,1); Inc(Depth);
     if (N and $3F) = $3F then FreshCounter;
     Ptr:=N; TINodeClear(A,V,Rt); Inc(N);
     BuildPart(A.L,Ptr,S,M);
     BuildPart(A.R,Ptr,M+1,E);
     PutRec0(Ptr,A);
     Dec(Depth); PV^.WriteChar(17+Depth,4,' ',1,1);
    end;
   end;
  end end;
 begin
  PDF:=aPDF; FldId:=aFldId; TIPartClear(Data); N:=0;
  FileDelete(aName);
  inherited InitHard(aName,SizeOf(TINode),StdIndxBufLen);
  F.Init(TmpIName,TmpISize); Depth:=0; GetMem(PB,TmpISize);
  BuildPart(Data,Root,0,F.Length);
  F.Done; FileDelete(TmpIName); FreeMem(PB,TmpISize);
  IndFlush;
 end;
procedure TIndex.ChangeName(aName:FullFName; aFldId:byte);
 var s:string[80]; f:file;
 begin
  s:=Name; inherited Done; assign(f,s); rename(f,aName);
  FldId:=aFldId; inherited Init(aName,SizeOf(TINode),StdIndxBufLen);
 end;
procedure TIndex.IndFlush;
 var A:TINode;
 begin
  {
   Ready to close file.
   If Cnt>1 then allwayse Data.Ptr=0 and A[0].B=Root !
  }
  if N>0
   then begin GetRec(0,A); A.B:=Data.Cnt end
   else TINodeClear(A,Data.Ptr,Data.Cnt);
  PutRec(0,A);
 end;
function TIndex.Less(X,Y:TValue):boolean;
 begin Less:=PDF^.Less(X,Y,FldId) end;
{ destructor TIndex.Done; begin inherited Done end; }
destructor TIndex.Kill;
 var s:string[80];
 begin 
  WrLogStr('IndKill start ...');
  s:=Name; 
  inherited Done;
  WrLogStr('IndKill Done OK ...');
  FileDelete(s) 
 end;
procedure TIndex.Show(P:TIPart);
 var i:word; A:TINode;
 begin with P do
  if Cnt=0 then write('.')
   else if Cnt=1 then write(Ptr)
    else begin
     write('(');
     GetRec(Ptr,A);
     with A do begin Show(L); write(' ',V,' '); Show(R) end;
     write(')');
     if Ptr=0 then begin
      writeln;
      for i:=0 to N-1 do begin
       GetRec(i,A);
       write(A.B mod 100:2,' ')
      end;
     end;
    end
 end;
procedure TIndex.Add(V:TValue);
 var i:longint; OK:boolean;
 begin
  i:=Find(V,OK);
  if not OK then begin GCnt:=0; AddPart(Data,Root,V); Garbage end;
 end;
procedure TIndex.ChB(P:TIPart; Root:longint);
 var A:TINode;
 begin with P do if Cnt>1 then begin
  GetRec(Ptr,A); A.B:=Root; PutRec0(Ptr,A)
 end end;
procedure TIndex.AddPart(var P:TIPart; Root:longint; aV:TValue);
 procedure Bal(var A:TINode; Ptr,Cnt:longint);
  var T:TValue; P1,C1:longint; W:TIPart; A1:TINode;
  begin
   with A do if L.Cnt<>Cnt then if L.Cnt>Cnt
    then if L.Cnt=1
     then begin AddPart(R,Ptr,V); V:=L.Ptr; L.Cnt:=0 end
     else begin
      P1:=L.Ptr; T:=V; C1:=R.Cnt;
      GetRec(P1,A1); Bal(A1,P1,Cnt);
      V:=A1.V; A1.V:=T; L:=A1.L; ChB(L,Ptr);
      A1.L:=A1.R; A1.R:=R;       ChB(R,P1);
      PutRec0(P1,A1);
      R.Ptr:=P1; R.Cnt:=1+A1.L.Cnt+C1;
      if R.Cnt=1 then begin R.Ptr:=A1.V; GclAdd(P1) end
     end
    else if R.Cnt=0
     then begin ErrNum(esBalInposibl); Halt end
     else if R.Cnt=1
      then begin AddPart(L,Ptr,V); V:=R.Ptr; R.Cnt:=0 end
      else begin
       P1:=R.Ptr; T:=V; C1:=L.Cnt;
       GetRec(P1,A1); Bal(A1,P1,Cnt-L.Cnt-1);
       V:=A1.V; A1.V:=T; R:=A1.R; ChB(R,Ptr);
       A1.R:=A1.L; A1.L:=L;       ChB(L,P1);
       PutRec0(P1,A1);
       L.Ptr:=P1; L.Cnt:=Cnt;
       if L.Cnt=1 then begin L.Ptr:=A1.V; GclAdd(P1) end
      end;
  end;
 var A:TINode;
 begin with P do begin
  inc(Cnt);
  if Cnt=1 then Ptr:=aV
   else begin
    if Cnt=2 then
     begin TINodeClear(A,Ptr,Root); PutRec0(N,A); Ptr:=N; inc(N) end;
    GetRec(Ptr,A);
    with A do begin
     if Less(aV,V) then AddPart(L,Ptr,aV) else AddPart(R,Ptr,aV);
     if (L.Cnt>=2*(R.Cnt+1)) or (R.Cnt>=2*(L.Cnt+1))
      then Bal(A,Ptr,Cnt div 2);
    end;
    PutRec0(Ptr,A);
   end;
 end end;
function TIndex.GetV(V:TPlace):TValue;
 begin if V<Data.Cnt
  then GetV:=GetVPart(Data,V)
  else GetV:=$7FFFFFFF
 end;
function TIndex.GetVPart(P:TIPart; aV:TPlace):TValue;
 var A:TINode;
 begin with P do if Cnt=1 then GetVPart:=Ptr else begin
  GetRec(Ptr,A);
  with A do if aV=L.Cnt
   then GetVPart:=V
   else if aV<L.Cnt
    then GetVPart:=GetVPart(L,aV)
    else GetVPart:=GetVPart(R,aV-L.Cnt-1)
 end end;
{ Mov is correct only for unique fields ! Don't use ! }
procedure TIndex.Mov(OV,NV:TValue);
 var Pl:TPlace; OK,WrOK:boolean;
 procedure SetNewV(var P:TIPart; aV:TPlace);
  var A:TINode;
  begin with P do if Cnt=1 then Ptr:=NV else begin
   GetRec(Ptr,A);
   with A do if aV=L.Cnt then V:=NV else if aV<L.Cnt
    then SetNewV(L,aV) else SetNewV(R,aV-L.Cnt-1);
   if not WrOK then begin PutRec(Ptr,A); WrOK:=true end
  end end;
 begin
  Pl:=Find(OV,OK); WrOK:=False;
  if OK
   then begin SetNewV(Data,Pl); if not WrOK then IndFlush end
   else ErrNum(esIndNotFound);
 end;
procedure TIndex.Del(V:TValue);
 var i:longint; OK:boolean;
 begin
  i:=Find(V,OK);
  if OK then begin GCnt:=0; DelPart(Data,V); Garbage end;
 end;
procedure TIndex.DelPart(var P:TIPart; aV:TValue);
 var T:TValue; A:TINode;
 begin with P do begin
  dec(Cnt);
  if Cnt>0 then begin
   GetRec(Ptr,A);
   if Cnt=1
    then with A do begin
     if aV=V then if L.Cnt=1 then T:=L.Ptr else T:=R.Ptr else T:=V;
     GclAdd(Ptr); Ptr:=T;
    end
    else begin
     with A do if aV=V
      then if L.Cnt>R.Cnt
       then begin T:=GetVPart(L,L.Cnt-1); V:=T; DelPart(L,T) end
       else begin T:=GetVPart(R,0); V:=T; DelPart(R,T) end
      else if Less(aV,V)
       then DelPart(L,aV)
       else DelPart(R,aV);
     PutRec0(Ptr,A);
    end;
  end;
 end end;
procedure TIndex.Garbage;
 var Rt,Pt:longint; i,j:byte; A,A1:TINode;
 begin
  if GCnt>0 then begin
   if GCnt>1 then for i:=0 to GCnt-2 do for j:=i+1 to GCnt-1 do
    if GArr[i]<GArr[j] then { swap & sort }
     begin Rt:=GArr[i]; GArr[i]:=GArr[j]; GArr[j]:=Rt end;
   for i:=0 to GCnt-1 do begin
    dec(N); GetRec(N,A); Rt:=A.B; Pt:=GArr[i];
    if N>Pt then begin
     PutRec0(Pt,A);
     GetRec(Rt,A1);
     with A1 do if ((L.Cnt>1) and (L.Ptr=N)) then L.Ptr:=Pt else R.Ptr:=Pt;
     PutRec0(Rt,A1);
     with A do begin ChB(L,Pt); ChB(R,Pt) end;
    end;
   end;
   Trunc0(N);
  end;
  IndFlush;
 end;
procedure TIndex.GclAdd(Pt:longint);
 begin
  if GCnt>GMax
   then begin ErrNum(esIndxStckOfl); Halt end
   else begin GArr[GCnt]:=Pt; inc(GCnt) end
 end;

function TIndex.Find(aV:TValue; var OK:boolean):TPlace;
 function FindPart(P:TIPart):TPlace;
  var A:TINode;
  begin with P do if Cnt<2
   then begin FindPart:=0; OK:=(aV=Ptr) and (Cnt=1) end
   else begin
    GetRec(Ptr,A);
    with A do if aV=V
     then begin OK:=True; FindPart:=L.Cnt end
     else if Less(aV,V)
      then FindPart:=FindPart(L)
      else FindPart:=1+L.Cnt+FindPart(R)
   end;
  end;
 begin OK:=False; Find:=FindPart(Data) end;
function TIndex.FindFirst(Dir:boolean):TPlace;
 function LP(l:longint):boolean;
  var OK:boolean;
  begin
   OK:=PDF^.LessP(l,FldId,Dir);
   if Dir then LP:=OK else LP:=not OK;
  end;
 function FindPart(P:TIPart):TPlace;
  var A:TINode;
  begin with P do if Cnt=0 then FindPart:=ord(Dir)-1
   else if Cnt=1 then
    if LP(Ptr) then FindPart:=ord(Dir)-1 else FindPart:=ord(Dir)
   else begin
    GetRec(Ptr,A);
    with A do if LP(V)
      then FindPart:=FindPart(L)
      else FindPart:=1+L.Cnt+FindPart(R)
   end;
  end;
 begin FindFirst:=FindPart(Data) end;

{---------------------------------------------------------}

var i:byte;

begin
 for i:=0 to 254 do DFArr[i]:=Nil;
 SystemMode:=InitMode;
end.
