unit DBF_def;

{$I options.pp }

interface

uses
 {$IFDEF Turbo} Skl_defs, {$ENDIF}
 {$IFDEF FPC  } Sk_types, {$ENDIF}
 Records;

const
 DBFNormalRec = ' ';
 DBFErasedRec = '*';

type
 TDBFHead = record
  { First 16 bytes: }
  DBF_Id:array[0..3] of byte;
  RecCount:longint;
  StartOfData:word;
  RecSize:word;
  Dummy0:array[0.. 3] of byte;
  { Second 16 bytes: }
  Dummy1:array[0..15] of byte;
 end;
 TDBFField = record
  { First 16 bytes: }
  FName:array[0..10] of byte;
  FType:char;
  Dummy0:array[0..3] of byte;
  { Second 16 bytes: }
  FSize:byte;
  DigAfterDot:byte;
  Dummy1:array[0..13] of byte;
 end;
 PDBFFile = ^TDBFFile;
 TDBFFile = object
  Head:TDBFHead;
  Flds:array[0..FldMaxCnt] of TDBFField; FldCnt:byte;
  f:file; PData:PBlock;
  sors_mode:byte; { 0-DOS 1-Win 2-Mac }
  constructor ImportDBF(FName:string);
  constructor ExportDBF(FName:string);
  procedure ImportFld(i:byte; var Fld:TFldRec);
  procedure ImportRec(var Rec; m:byte);
  destructor Done;
 end;

{
  DBF record structure:
  RecCase:char; Fld1, Fld2, ... FldN

  DBF file structure:
  DBFHead,
  Field1, Field2, ..., FieldN,
  CR,
  Rec1, Rec2, ... RecM

}

implementation

uses
 date_lib, win_def;

procedure TDBFHeadInit(var DBH:TDBFHead; aRC:longint; aFldCnt:byte; aRS:word);
 begin with DBH do begin
  DBF_Id[0]:=$03; DBF_Id[1]:=$61; DBF_Id[2]:=$05; DBF_Id[3]:=$0D;
  RecCount:=aRC;
  StartOfData:=(aFldCnt+1)*32+1;
  RecSize:=aRS;
  FillChar(Dummy0,SizeOf(Dummy0),chr(0));
  FillChar(Dummy1,SizeOf(Dummy1),chr(0));
 end end;

procedure TDBFFieldInit(
 var DBF:TDBFField; aName:string; aFT:char; aFS,aDAD:byte);
 var s:string;
 begin with DBF do begin
  s:=copy(aName,1,10);
  while length(s)<11 do s:=s+chr(0);
  { Name must contain only UpCase ASCII leters in range [$B0..$5F] }
  StrRec(s,FName,SizeOf(FName));
  FType:=aFT;
  DigAfterDot:=0;
  case aFT of
   'D':FSize:=8;
   'C':FSize:=aFS;
   'N':begin FSize:=aFS; DigAfterDot:=aDAD end;
  end;
  FillChar(Dummy0,SizeOf(Dummy0),chr(0));
  FillChar(Dummy1,SizeOf(Dummy1),chr(0));
 end end;
function TDBFFieldNumLongOK(var DBF:TDBFField):boolean;
 begin with DBF do TDBFFieldNumLongOK:=(DigAfterDot=0) and (FSize<10) end;

{------------------------------------------------------------------}

constructor TDBFFile.ImportDBF(FName:string);
 var i:byte; PCL:PChoisList;
 begin
  assign(f,FName); reset(f,1);
  BlockRead(f,Head,SizeOf(Head));
  FldCnt:=((Head.StartOfData-1) div 32) - 1;
  if FldCnt>0 then for i:=0 to min(FldMaxCnt,FldCnt-1) do
   BlockRead(f,Flds[i],SizeOf(TDBFField));
  Seek(f,Head.StartOfData);
  GetMem(PData,Head.RecSize);
  New(PCL,Init);
  PCL^.Insert('     (DOS ) ');
  PCL^.Insert('    Windows -> DOS  ');
  PCL^.Insert('    Macintosh -> DOS ');
  {$IFDEF Turbo} Dispose(PCL,Chois(sors_mode)); {$ENDIF}
  {$IFDEF FPC} PCL^.Chois(sors_mode); {$ENDIF}
  if Sors_mode>2 then Sors_mode:=0;
 end;
procedure TDBFFile.ImportFld(i:byte; var Fld:TFldRec);
 var s:string;
 begin with Fld do begin
  FldClear(Fld,i); s:=RecStr(Flds[i].FName,11);
  StrRec(copy(s,1,pos(chr(0),s)-1),Name,FldMaxLen);
  Size:=Flds[i].FSize; ShowSize:=0;
  case Flds[i].FType of
   'N':begin
    ShowSize:=Size;
    if TDBFFieldNumLongOK(Flds[i])
     then begin FType:=CFLong; Size:=SizeOf(longint) end
     else begin FType:=CFReal; Size:=SizeOf(real) end;
   end;
   'D':begin FType:=CFDate; Size:=SizeOf(DateRec) end;
   else FType:=CFStr; { 'C' }
  end;
  if ShowSize=0 then ShowSize:=FldStdStrLen(Fld);
  Flags:=0; if i=0 then FldSetFlags(Fld,ffIndexed + ffRootInd,True);
  StrRec('',CntName,6);
 end end;
procedure TDBFFile.ImportRec(var Rec; m:byte);
 var R:Block absolute Rec; c0,c1:word; i:byte;
 procedure ImpLong(l:byte);
  var s:string; w:longint; i:integer;
  begin
   s:=ClearSp(RecStr(PData^[c0],l));
   Val(s,w,i); if i<>0 then w:=0;
   Move(w,R[c1],SizeOf(longint));
  end;
 procedure ImpReal(l:byte);
  var s:string; w:real; i:integer;
  begin
   s:=ClearSp(RecStr(PData^[c0],l));
   Val(s,w,i); if i<>0 then w:=0.0;
   Move(w,R[c1],SizeOf(real));
  end;
 procedure ImpDate;
  var s:string[4]; d0:DateRec; i,w:integer;
  procedure rd2(x:byte);
   begin
    s:=ClearSp(RecStr(PData^[c0+x],2));
    if (length(s)=2) and (s[1]='0') then s:=s[2];
    Val(s,w,i); if i<>0 then w:=0;
   end;
  begin
   s:=ClearSp(RecStr(PData^[c0],4));
   Val(s,w,i); if i<>0 then w:=1900;
   if (w>=1900) and (w<2150) then d0.y:=w-1900 else d0.y:=0;
   rd2(4); if (w>=1) and (w<13) then d0.m:=w else d0.m:=0;
   rd2(6); if (w>=1) and (w<32) then d0.d:=w else d0.d:=0;
   Move(d0,R[c1],SizeOf(d0));
  end;
 procedure ConvertStr(var R; Size:byte);
  var i:byte;
  begin case sors_mode of
   0{DOS}:;
   1{Win}:for i:=0 to Size-1 do Small(R)[i]:=WinDosChar(Small(R)[i]);
   2{Mac}:for i:=0 to Size-1 do Small(R)[i]:=MacDosChar(Small(R)[i]);
  end end;
 begin
  BlockRead(f,PData^,Head.RecSize);
  case chr(PData^[0]) of
   DBFNormalRec:R[0]:=0;
   DBFErasedRec:R[0]:=$FF;
  end;
  c0:=1; c1:=1;
  for i:=0 to m do with Flds[i] do begin
   case FType of
    'N':if TDBFFieldNumLongOK(Flds[i])
         then begin ImpLong(FSize); Inc(c1,SizeOf(longint)) end
         else begin ImpReal(FSize); Inc(c1,SizeOf(real)) end;
    'D':begin ImpDate; Inc(c1,SizeOf(DateRec)) end;
    else begin
     Move(PData^[c0],R[c1],FSize);
     ConvertStr(R[c1],FSize);
     Inc(c1,FSize)
    end; { 'C' }
   end;
   inc(c0,FSize);
  end;
 end;
destructor TDBFFile.Done;
 begin
  FreeMem(PData,Head.RecSize);
  close(f);
 end;

constructor TDBFFile.ExportDBF(FName:string);
 begin
 end;

end.
