unit Disk_rec;

{ Access to disk files, containing records }
{ (c) Skelet Software, 1993, 1998          }

{$I options.pp}

interface

uses
 {$IFDEF FPC}
 sk_types, unl_fils,
 {$ENDIF}
 {$IFDEF Turbo}
 skl_defs, files,
 {$ENDIF}
 mem_rec;

type

 TRecFile = object
  constructor Init(aName:StdString; aRecSize:Long);
  function Name:StdString;
  function Length:Long;
  function Active:boolean;
  destructor Done;
 private
  FL:TVirtFile;
  TRF_RecSize:Long;
  LenVar:Long;
 end;

 TRecFileDrv = object (TRecFile)
  procedure GetRec(ind:Long; var rec); virtual;
  procedure PutRec0(ind:Long; var rec); virtual;
  procedure PutRec(ind:Long; var rec); virtual;
  procedure Trunc0(ind:Long); virtual;
  function  Append0(var rec):Long; virtual;
  function  Append (var rec):Long; virtual;
  procedure Flush;                 virtual;
  function  Size:Long;
  procedure FClose;
  procedure FOpen;
{$ifdef Turbo} private {$endif}
  function  TheSame(ind:Long; var rec):boolean; virtual;
{$ifdef FPC} private {$endif}
  procedure ReadTest(ind:Long);
  procedure DiskRead(ind:Long; var rec);
  procedure DiskWrite(ind:Long; var rec);
  procedure DiskFlush;
 end;

 PRecFileCashDrv = ^TRecFileCashDrv;
 TRecFileCashDrv = object (TRecFileDrv)
  constructor Init(
   aName:StdString;
   aRecSize:Long;
   aBufLen:word);  { Cash buffer size in records }
  constructor InitHard(
   aName:StdString;
   aRecSize:Long;
   aBufLen:word);  { Cash buffer size in records }
  procedure  GetRec(ind:Long; var rec); virtual;
  procedure  Trunc0(ind:Long); virtual;
  function   BuferLen:word;
  destructor Done;
{$ifdef FPC} 
  function TheSame(ind:Long; var rec):boolean; virtual;
{$endif}
 private
  Cash:TRecCash;
  BufLen:word;
{$ifdef Turbo} 
  function TheSame(ind:Long; var rec):boolean; virtual;
{$endif}
 end;

 TRecFileFast = object (TRecFile)
  buf:PBlock;
  constructor Init(aName:StdString; aRecSize:Long);
  destructor Done;
 private
  bufsize,k,j: Long;
 end;
 TRecFileFastIn = object (TRecFileFast)
  e_file:boolean;
  constructor Init(aName:StdString; aRecSize:Long);
  procedure  GetNext(var rec);
  destructor Kill;
 private
  l:Long;
  procedure GetBlock;
 end;
 TRecFileFastOut = object (TRecFileFast)
  constructor Init(aName:StdString; aRecSize:Long);
  procedure  PutNext(var rec);
  destructor Done;
 end;

implementation

uses
 Trace;
 
constructor TRecFileCashDrv.InitHard(
 aName:StdString;
 aRecSize:Long;
 aBufLen:word);
 begin
  inherited Init(aName,aRecSize);
  BufLen:=aBufLen;
  Cash.Init(TRF_RecSize,BufLen);
 end;
constructor TRecFileCashDrv.Init(
 aName:StdString;
 aRecSize:Long;
 aBufLen:word);
 begin
  inherited Init(aName,aRecSize);
  if (Pos('.I',aName)>0) or (Pos('.i',aName)>0) { Index files }
   then BufLen:=min(aBufLen,Length div 3 + 16)
   else BufLen:=min(aBufLen,Length+5);
  Cash.Init(TRF_RecSize,BufLen);
 end;
procedure TRecFileCashDrv.GetRec(ind:Long; var rec);
 var ok:boolean;
 begin
  ReadTest(ind); Cash.ReadRec(ind,rec,ok);
  if not ok then begin 
{  WrLogStr('GetRec ind=$'+HexLong(ind)+' RecSize=$'+HexLong(TRF_RecSize)); }
   DiskRead(ind,rec);
   Cash.Mark(ind,rec)
  end;
 end;
procedure TRecFileCashDrv.Trunc0(ind:Long);
 begin inherited Trunc0(ind); Cash.Trunc(ind) end;
function TRecFileCashDrv.BuferLen:word; begin BuferLen:=BufLen end;
function TRecFileCashDrv.TheSame(ind:Long; var rec):boolean;
 var ok:boolean; begin Cash.WriteRec(ind,rec,ok); TheSame:=ok end;
destructor TRecFileCashDrv.Done;
 begin Cash.Done; inherited Done end;

{---------------------------------------------------------}

function  TRecFileDrv.TheSame(ind:Long; var rec):boolean;
 begin TheSame:=False end;
procedure TRecFileDrv.ReadTest(ind:Long);
 begin
  if (Ind>=LenVar) or (Ind<0) then begin
   writeln('Attempt to read record # ',Ind);
   StdError('DiskRead out of range for file '+Name)
  end;
 end;
procedure TRecFileDrv.DiskRead(ind:Long; var rec);
 begin FL.FSeek(ind*TRF_RecSize); FL.BRead(rec,TRF_RecSize) end;
procedure TRecFileDrv.DiskWrite(ind:Long; var rec);
 begin
  FL.FSeek(ind*TRF_RecSize); FL.BWrite(rec,TRF_RecSize);
  if ind>=LenVar then LenVar:=Ind+1;
 end;
procedure TRecFileDrv.Trunc0(ind:Long);
 begin FL.FTrunc(ind*TRF_RecSize); LenVar:=Ind end;
procedure TRecFileDrv.DiskFlush; begin FL.FFlush end;
procedure TRecFileDrv.GetRec(ind:Long; var rec);
 begin ReadTest(ind); DiskRead(ind,rec) end;
procedure TRecFileDrv.PutRec0(ind:Long; var rec);
 begin if not TheSame(ind,rec) then DiskWrite(ind,rec) end;
procedure TRecFileDrv.PutRec(ind:Long; var rec);
 begin if not TheSame(ind,rec) then begin DiskWrite(ind,rec); DiskFlush end end;
function  TRecFileDrv.Append0(var Rec):Long;
 var Ind:Long; begin Ind:=Length; PutRec0(Ind,Rec); Append0:=Ind end;
function  TRecFileDrv.Append(var Rec):Long; begin Append:=Append0(Rec); Flush end;
procedure TRecFileDrv.Flush; begin DiskFlush end;
function  TRecFileDrv.Size:Long; begin Size:=TRF_RecSize end;
procedure TRecFileDrv.FClose; begin FL.FClose end;
procedure TRecFileDrv.FOpen;  begin FL.FOpen end;

{---------------------------------------------------------}

const
 MaxBuffer={$IFDEF FPC} 200000 {$ELSE} 65000 {$ENDIF};

constructor TRecFileFast.Init(aName:StdString; aRecSize:Long);
 begin
  inherited Init(aName,aRecSize);
  BufSize:=min(MaxBuffer,MaxAvail div 3);
  k:=(bufsize div TRF_RecSize) * TRF_RecSize;
  BufSize:=k; GetMem(buf,bufsize);
 end;
destructor TRecFileFast.Done;
 begin Inherited Done; FreeMem(buf,bufsize) end;

procedure TRecFileFastIn.GetBlock;
 begin
  e_file:=l=0; j:=0;
  if l>k then Dec(l,k) else begin k:=l; l:=0 end;
  if k>0 then FL.BRead(buf^,k);
 end;
constructor TRecFileFastIn.Init(aName:StdString; aRecSize:Long);
 begin
  inherited Init(aName,aRecSize);
  l:=Length * TRF_RecSize; GetBlock;
 end;
procedure TRecFileFastIn.GetNext(var rec);
 begin
  Move(buf^[j],rec,TRF_RecSize);
  Inc(j,TRF_RecSize); if j=k then GetBlock;
 end;
destructor TRecFileFastIn.Kill;
 var s:StdString;
 begin s:=FL.FName; inherited Done; FileDelete(s) end;

constructor TRecFileFastOut.Init(aName:StdString; aRecSize:Long);
 begin FileDelete(aName); inherited Init(aName,aRecSize); j:=0 end;
procedure  TRecFileFastOut.PutNext(var rec);
 begin
  Move(rec,buf^[j],TRF_RecSize); Inc(j,TRF_RecSize);
  if j=k then begin FL.BWrite(buf^,k); j:=0 end;
  Inc(LenVar);
 end;
destructor TRecFileFastOut.Done;
 begin if j>0 then FL.BWrite(buf^,j); inherited Done end;

constructor TRecFile.Init(aName:StdString; aRecSize:Long);
 var fs:Long;
 begin
  TRF_RecSize:=aRecSize; FL.Init(aName); fs:=FL.FSize;
  if (fs mod aRecSize)>0
   then StdError('(FileSize mod TRF_RecSize)>0 for file '+aName)
   else LenVar:=fs div aRecSize;
 end;
function TRecFile.Length:Long; begin Length:=LenVar end;
function TRecFile.Active:boolean; begin Active:=FL.FActive end;
function TRecFile.Name:StdString; begin Name:=FL.FName end;
destructor TRecFile.Done;
 begin FL.Done end;

end.
