unit mem_rec;

interface

uses sk_types;

type

 TRecCashLink = object
  constructor Init(aLen:word);
  procedure Search (aInd:Long; var i:word; var ok:boolean);
  procedure IncStat(i:word);
  procedure NewStat(NewInd:Long; var i:word);
  procedure ChLinks(i,NewInd:Long);
  destructor Done;
 private
  Len:word;
  Tab,Ptb:PBlock;
  CurTime:byte;
 end;

 TRecCash = object
  constructor Init(aRecSize:Long; var BufLen:word);
  procedure  Mark   (ind:Long; var rec);
  procedure  ReadRec(ind:Long; var rec; var ok:boolean);
  function   GetByte(ind1,ind2:Long; var ok:boolean):byte;
  procedure  WriteRec(ind:Long; var rec; var ok:boolean);
  procedure  Trunc(ind:Long);
  destructor Done;
 private
  RecSize:Long; Buf:PBlock;
  RCL:TRecCashLink;
  procedure NewStat(NewInd:Long; var rec);
 end;

implementation

const
 MaxFileIndex = 1000000000;

{--------------------------------------------------------}

type
 TCInd = record ind:Long; ptr:word; density,time:byte end;
 TCArr = array[0..5000] of TCInd;
 TIArr = array[0..32000] of word;
const
 TbSize=SizeOf(TCInd);

{--------------------------------------------------------}

constructor TRecCash.Init(aRecSize:Long; var BufLen:word);
 var UsedMem:Long; BL,MaxBufSz,RUMem:Long; OK:boolean;
 begin
  RUMem:=TbSize+SizeOf(word){ RCL }+aRecSize{ Self };
  UsedMem:=Long(BufLen)*RUMem;
  MaxBufSz:=min(max(0,MaxAvail-2000),40000);
  if UsedMem>MaxBufSz then begin
   BL:=MaxBufSz div RUMem;
   OK:=BL>0; if OK then OK:=(BufLen div BL) < 10;
   if OK then BufLen:=BL else StdError('No memory for cash buffers');
  end;
  RecSize:=aRecSize;
  GetMem(Buf,Long(BufLen)*RecSize);
  RCL.Init(BufLen);
 end;
procedure TRecCash.Mark(ind:Long; var rec);
 var
  found:boolean; i:word;
 begin
  RCL.Search(ind,i,found);
  if found then RCL.IncStat(i) else NewStat(ind,rec);
 end;
procedure TRecCash.ReadRec(ind:Long; var rec; var ok:boolean);
 var i:word;
 begin
  RCL.Search(ind,i,ok);
  if ok then begin
   Move(Buf^[RecSize*i],rec,RecSize); RCL.IncStat(i) end;
 end;
function TRecCash.GetByte(ind1,ind2:Long; var ok:boolean):byte;
 var i:word;
 begin
  RCL.Search(ind1,i,ok);
  if ok then begin GetByte:=Buf^[RecSize*i+ind2]; RCL.IncStat(i) end;
 end;
procedure TRecCash.WriteRec(ind:Long; var rec; var ok:boolean);
 var i:word;
 begin
  RCL.Search(ind,i,ok);
  if ok
   then begin
    if not equ(Buf^[RecSize*i],rec,RecSize)
     then begin
      Move(rec,Buf^[RecSize*i],RecSize); ok:=false end;
    RCL.IncStat(i);
   end
   else NewStat(ind,rec);
 end;
procedure  TRecCash.Trunc(ind:Long);
 var i:word; ok:boolean;
 begin
  RCL.Search(ind,i,ok);
  if ok then RCL.ChLinks(i,MaxFileIndex);
 end;
destructor TRecCash.Done;
 begin
  FreeMem(Buf,Long(RCL.Len)*RecSize);
  RCL.Done;
 end;
procedure TRecCash.NewStat(NewInd:Long; var rec);
 var i:word;
 begin
  RCL.NewStat(NewInd,i);
  Move(rec,Buf^[RecSize*i],RecSize);
 end;

{--------------------------------------------------------}

constructor TRecCashLink.Init(aLen:word);
 var
  Tarr:^TCArr absolute Tab;
  Iarr:^TIArr absolute Ptb;
  r:TCInd; i:word;
 begin
  Len:=aLen;
  GetMem(Tab,Len*TbSize);
  GetMem(Ptb,Len*SizeOf(word));
  CurTime:=1;
  r.ind:=MaxFileIndex; { Unused index }
  r.density:=0;  { Never read   }
  r.time:=0;     { Not read }
  r.ptr:=Len;    { Nil ptr }
  for i:=0 to Len-1 do begin
   Tarr^[i]:=r;
   r.ptr:=i; Iarr^[i]:=Len
  end;
  Iarr^[MaxFileIndex mod Len]:=Len-1; { Link all blocks }
 end;
procedure TRecCashLink.ChLinks(i,NewInd:Long);
 var
  Tarr:^TCArr absolute Tab;
  Iarr:^TIArr absolute Ptb;
  m,mi:word;
 begin
  m:=Tarr^[i].ind mod Len;
  mi:=Iarr^[m];                      { Entry in Iarr }
  if mi=i
   then Iarr^[m]:=Tarr^[mi].ptr
   else begin
    while mi<>i do begin m:=mi; mi:=Tarr^[m].ptr end;
    Tarr^[m].ptr:=Tarr^[mi].ptr
   end;                              { Old link cleared }
  with Tarr^[i] do begin
   ind:=NewInd; density:=0; time:=0; { New Statistics OK }
   m:=ind mod Len; ptr:=Iarr^[m];
  end;
  Iarr^[m]:=i;                       { New link OK }
 end;
procedure TRecCashLink.NewStat(NewInd:Long; var i:word);
 var
  Tarr:^TCArr absolute Tab;
  m,mi:word;
 begin
  { Find min: }
  m:=65000;
  for mi:=0 to Len-1 do with Tarr^[mi] do
   if m>density+time then begin m:=density+time; i:=mi end;
  ChLinks(i,NewInd); IncStat(i);
 end;
procedure TRecCashLink.Search(aInd:Long; var i:word;var ok:boolean);
 var
  Tarr:^TCArr absolute Tab;
  Iarr:^TIArr absolute Ptb;
 begin
  ok:=false; i:=Iarr^[aInd mod Len];
  while (i<Len) and (not ok) do with Tarr^[i] do
   if aInd=ind then ok:=true else i:=ptr;
 end;
procedure TRecCashLink.IncStat(i:word);
 var
  Tarr:^TCArr absolute Tab; w:word;
 begin
  with Tarr^[i] do begin
   if density=255 then for w:=0 to Len-1 do
    Tarr^[w].density:=Tarr^[w].density div 2;
   inc(density); time:=CurTime;
  end;
  if CurTime=255 then begin
   for w:=0 to Len-1 do Tarr^[w].time:=Tarr^[w].time div 2;
   CurTime:=127;
  end;
  inc(CurTime);
 end;
destructor TRecCashLink.Done;
 begin
  FreeMem(Tab,Len*TbSize);
  FreeMem(Ptb,Len*SizeOf(word));
 end;

{$R-}

end.


