unit records;
interface

{$I options.pp}

uses
 {$IFDEF Turbo} Skl_defs, {$ENDIF}
 {$IFDEF FPC  } Sk_types, {$ENDIF}
 Date_lib, tv_box, Views, Drivers, Objects;

{------------- Field definitions: --------------}

const
 { Field Type constants: }
 CFList = 0; CFWord = 1; CFLong = 2; CFPntr = 3;
 CFReal = 4; CFStr  = 5; CFName = 6; CFDate = 7;
 CFTime = 8; CFMax  = CFTime;

{
    FType       Size        FPtr      StdLayout

    byte         1          ---       Lists, Boolean
    word         2          ---       Number16, Integer16
    Ptr          4          File#
    String       n          ---
    Name         n          ---
    Date         3          ---       DateStr
    Time         3          ---       TymeStr
    Real         6          ---       Money, Real
}

const
 FldMaxLen = 20;
 FldMaxCnt = 20;

const
 { Field flags constants: }
 ffIndexed = $01;
 ffUniTest = $02;
 ffRootInd = $04;
 ffFulShow = $08;
 ffCntSumm = $10;
 ffCntPntr = $20;

type
 PFldRec = ^TFldRec;
 TFldRec = record
  Used: byte; { 0 - active, $FF - erased }
  Name: array[0..FldMaxLen-1] of byte;
  FType: byte;
  Size: byte;
  FPtr: byte; { if FType=Ptr, number for pointed file }
  Flags:byte; { Indexed, UniqueTest, RootIndex, FullShow, CountSumm, CountPntr }
  FPName: SmallFName; { Name for pointed file }
  ShowSize:byte;      { column size in Brouse window }
  CntName:array[0..5] of byte; { links counter name }
 end;

procedure FldSetFlags(var F:TFldRec; aFlags:byte; Enable:boolean);
function  FldGetFlags(var F:TFldRec; aFlags:byte):boolean;
function FldLess(var F:TFldRec; var R1,R2):boolean;
function FldEqual(var F:TFldRec; var R1,R2):boolean;
function FldLimitStr(var F:TFldRec; var R):string;
procedure FldChangeLimits(var F:TFldRec; var R; Var Ltd:boolean);
procedure FldClear(var F:TFldRec; i:byte);
function  FldStdStrLen(var F:TFldRec):byte;
function  FldStdLinLen(var F:TFldRec):byte;
function  FldStdString(var F:TFldRec; var R; Depth:byte):string;
function  FldStdBrouseStr(var F:TFldRec; var R):string;
procedure FldStdClear(var F:TFldRec; var R);


function FieldStr(Fld:TFldRec):string;
function FldTStr(FT:byte):string;
function LoRecStr(var R; l:byte):string;

{--------------- Input layouts: ----------------}

type
 PSRecDialog = ^TSRecDialog;
 TSRecDialog = object (TSDialog)
  VCnt:byte;
  CArr:Array[0..FldMaxCnt] of PView;
  bOK, bCancel:PView;
  constructor Init(var Bounds: TRect; ATitle: TTitleStr);
  procedure HandleEvent(var Event: TEvent); virtual;
  procedure AddNewView(ALabel:string; aPV:PView); virtual;
  procedure AddNewLine(ALabel:string; aML:byte; cs:boolean); virtual;
  procedure AddNewStr (ALabel:string; aML:byte; cs:boolean); virtual;
  procedure AddNewByte(ALabel:string); virtual;
  procedure AddBoolNum(Num:word); virtual;
  procedure AddStrNum (Num:word; aML:byte); virtual;
  procedure AddOKCancelButtons; virtual;
  procedure AddLayoutLine(
   FldT,FSize,SSize,FPtr:byte; var R; Tit:string); virtual;
 end;

{-----------------------------------------------}

type
 TPeriod = record
  D0,D1:DateRec;
 end;

procedure EditSomeString0(Tit:string; var S:string; Len:byte; Cs:boolean);
procedure EditSomeString(Tit:string; var S:string; Len:byte);
procedure EditSomeName(Tit:string; var S:string; Len:byte);
procedure EditPeriod(var Prd:TPeriod; var OK:boolean);
function  EditDate(var D:DateRec):boolean;
function  EditSomePntr(FId:byte; var Pntr:longint):boolean;
procedure EditFirstFldValue(Fld:TFldRec; var Rec; aCase:byte);
procedure LepkaNumEdit(FId:byte; var Sors,Dest:longint);

var
 EditStringOK:boolean;

{-----------------------------------------------}

function InputLayout(var Rec; FId:byte;
 aInd:longint; abwCase,aExFld:byte):boolean;

{-----------------------------------------------}

implementation

uses
 Menu_Cmd, Brousers, DataFile, Win_Def, Lines, Str_def,
 App, Dialogs;

{-------------- Field definition: --------------}

function LoCaseB(b:byte):byte;
 const BigLSet = ['A'..'Z',''..''];
 begin LoCaseB:=b+(ord(chr(b) in BigLSet) shl 5) end;
function LoRecStr(var R; l:byte):string;
 var s:string; i:byte;
 begin
  s:=ClearSp(RecStr(R,l));
  if length(s)>0 then for i:=1 to Length(s)
   do s[i]:=chr(LoCaseB(ord(s[i])));
  LoRecStr:=s;
 end;

procedure FldSetFlags(var F:TFldRec; aFlags:byte; Enable:boolean);
 begin with F do if Enable
   then Flags:=Flags or aFlags
   else Flags:=Flags and (aFlags xor $FF)
 end;
function  FldGetFlags(var F:TFldRec; aFlags:byte):boolean;
 begin with F do FldGetFlags:=(Flags and aFlags)=aFlags end;
function FldLess(var F:TFldRec; var R1,R2):boolean;
 var
  b1:byte absolute R1;    b2:byte absolute R2;
  w1:word absolute R1;    w2:word absolute R2;
  l1:longint absolute R1; l2:longint absolute R2;
  rr1:real absolute R1;   rr2:real absolute R2;
  d1:DateRec absolute R1; d2:DateRec absolute R2;
  t1:TimeRec absolute R1; t2:TimeRec absolute R2;
  sr1:small absolute R1;  sr2:small absolute R2; i:byte; Eq:boolean;
 begin with F do case FType of
  CFList:FldLess:=b1<=b2;
  CFWord:FldLess:=w1<=w2;
  CFLong,CFPntr:FldLess:=l1<=l2;
  CFReal:FldLess:=rr1<=rr2;
  CFStr,CFName:begin
   Eq:=True; i:=0;
   while Eq and (i<Size)
    do begin Eq:=LoCaseB(sr1[i])=LoCaseB(sr2[i]); inc(i) end;
   dec(i);
   if not Eq then Eq:=LoCaseB(sr1[i])<LoCaseB(sr2[i]);
   FldLess:=Eq;
  end;
  CFDate:FldLess:=DateLess(d1,d2);
  CFTime:FldLess:=TimeLess(t1,t2);
 end end;
function FldEqual(var F:TFldRec; var R1,R2):boolean;
 var sr1:small absolute R1; sr2:small absolute R2; i:byte; Eq:boolean;
 begin with F do case FType of
  CFStr, CFName:begin
   Eq:=True; i:=0;
   while Eq and (i<Size) do
    begin Eq:=LoCaseB(sr1[i])=LoCaseB(sr2[i]); inc(i) end;
   FldEqual:=Eq;
  end;
  else FldEqual:=Equ(R1,R2,Size);
 end end;
function FldLimitStr(var F:TFldRec; var R):string;
 var ps:PString; sn:string[20];
  Pntr:longint absolute R;
  Period:TPeriod absolute R;
  Num:array[0..1] of longint absolute R;
 begin with F do begin
  New(ps); ps^:='';
  case FType of
   CFList,CFWord,CFLong:begin
           Str(Num[0],ps^);
           if Num[1]>Num[0]
{$IFDEF Cir}
            then begin Str(Num[1],sn); ps^:=' '+ps^+'..'+sn end
            else ps^:='  '+ps^;
{$ENDIF}
{$IFDEF Lat}
            then begin Str(Num[1],sn); ps^:='interval '+ps^+'..'+sn end
            else ps^:='equal to '+ps^;
{$ENDIF}
          end;
   CFPntr:ps^:=DFArr[FPtr]^.StdPtrLine(Pntr,3);
   CFDate:with Period do ps^:=DateStr_2(D0)+' .. '+DateStr_2(D1);
   CFStr,CFName:
{$IFDEF Cir}
          ps^:=': '+ClearSp(RecStr(R,Size))+'@';
{$ENDIF}
{$IFDEF Lat}
          ps^:='Pattern: '+ClearSp(RecStr(R,Size))+'@';
{$ENDIF}
  end;
  FldLimitStr:='('+ps^+')';
  Dispose(ps);
 end end;
procedure FldChangeLimits(var F:TFldRec; var R; Var Ltd:boolean);
 var s:string; sn:string[20]; i,j:integer;
  Period:TPeriod absolute R;
  Pntr:longint absolute R;
  Num:array[0..1] of longint absolute R;
 procedure ConvertNum;
  begin if Pos('..',s)>1
   then begin
    j:=Pos('..',s);
    Val(copy(s,1,j-1),Num[0],i);
    Val(copy(s,j+2,length(s)-j-1),Num[1],i);
    Num[1]:=max(Num[1],Num[0]);
   end
   else begin Val(s,Num[0],i); Num[1]:=Num[0] end;
  end;
 begin with F do case FType of
  CFList,CFWord,CFLong:begin
   if not Ltd then begin Num[0]:=0; Num[1]:=100 end;
   Str(Num[0],s);
   if Num[1]>Num[0] then begin Str(Num[1],sn); s:=s+'..'+sn end;
   EditSomeString(
    ErrStr(esIntervOrVal)+
    ClearSp(RecStr(Name,FldMaxLen))+':',s,12);
   ConvertNum; Ltd:=True;
  end;
  CFPntr:begin
    if not Ltd then Pntr:=-1;
    if EditSomePntr(FPtr,Pntr) then Ltd:=True;
   end;
  CFDate:begin
    if not Ltd then with Period do
     begin DateGetD(D0); D1:=D0; D0.d:=1; D0.m:=1 end;
    EditPeriod(Period,Ltd);
   end;
  CFStr,CFName:begin
   if not Ltd then StrRec('',R,Size); s:=ClearSp(RecStr(R,Size));
{$IFDEF Cir}
   EditSomeString0('  '+ClearSp(RecStr(Name,FldMaxLen))+':',
{$ENDIF}
{$IFDEF Lat}
   EditSomeString0('Start leters for '+ClearSp(RecStr(Name,FldMaxLen))+':',
{$ENDIF}
                    s,Size,FType=CFStr);
   StrRec(s,R,Size); Ltd:=s<>'';
  end;
 end end;
procedure FldClear(var F:TFldRec; i:byte);
 var s:string[3];
 begin with F do begin
  Used:=0;  Str(i,s);
{$IFDEF Cir}
  StrRec('_'+s,Name,FldMaxLen);
{$ENDIF}
{$IFDEF Lat}
  StrRec('Field_'+s,Name,FldMaxLen);
{$ENDIF}
  FType:=CFStr; Size:=5; FPtr:=$FF;
  FPName:='';
  ShowSize:=FldStdStrLen(F);
  Flags:=0;
  Str(i,s);
{$IFDEF Cir}
  StrRec(''+s,CntName,6);
{$ENDIF}
{$IFDEF Lat}
  StrRec('Cnt_'+s,CntName,6);
{$ENDIF}
 end end;
function  FldStdStrLen(var F:TFldRec):byte;
 begin with F do case FType of
  CFList:FldStdStrLen:=5;
  CFWord:FldStdStrLen:=6;
  CFLong:FldStdStrLen:=11;
  CFPntr:FldStdStrLen:=15;
  CFReal:FldStdStrLen:=12;
  CFStr, CFName:FldStdStrLen:=Size;
  CFDate,CFTime:FldStdStrLen:=8;
 end end;
function  FldStdLinLen(var F:TFldRec):byte;
 begin with F do case FType of
  CFList:FldStdLinLen:=5;
  CFWord:FldStdLinLen:=7;
  CFLong:FldStdLinLen:=14;
  CFPntr:FldStdLinLen:=50;
  CFReal:FldStdLinLen:=14;
  CFStr, CFName:FldStdLinLen:=Size+2;
  CFDate:if ShowSize>8
   then FldStdLinLen:=10 else FldStdLinLen:=8;
  CFTime:FldStdLinLen:=8;
 end end;
function  FldStdString(var F:TFldRec; var R; Depth:byte):string;
 var ps:PString;
  b:byte absolute R;
  w:word absolute R;
  l:longint absolute R;
  rr:real absolute R;
  sr:small absolute R;
  d:DateRec absolute R;
  t:TimeRec absolute R;
 begin
  New(ps);
  with F do case FType of
   CFList:Str(b,ps^);
   CFWord:Str(w,ps^);
   CFLong:Str(l,ps^);
   CFPntr:if l>=0
          then ps^:=DFArr[FPtr]^.StdPtrLine(l,Depth)
          else ps^:='-';
   CFReal:Str(rr:12:2,ps^);
   CFStr, CFName:ps^:=RecStr(sr,Size);
   CFDate:case ShowSize of
            0,8:ps^:=DateStr_1(d);
           10:ps^:=DateStr_2(d);
          end;
   CFTime:ps^:=TimeStr_1(t);
  end;
  FldStdString:=ps^;
  Dispose(ps);
 end;
function  FldStdBrouseStr(var F:TFldRec; var R):string;
 var ps:PString;
 begin
  New(ps); ps^:=ClearSp(FldStdString(F,R,3));
  with F do case FType of
   CFPntr,CFStr,CFName:ps^:=JustStr(ps^,ShowSize);
   CFList,CFWord,CFLong,CFReal:if Length(ps^)>ShowSize
    then begin ps^:=''; LExpandStr(ps^,ShowSize,'#') end
    else LExpandStr(ps^,ShowSize,' ');
  end;
  FldStdBrouseStr:=ps^; Dispose(ps);
 end;
procedure FldStdClear(var F:TFldRec; var R);
 var
  b:byte absolute R;
  w:word absolute R;
  l:longint absolute R;
  rr:real absolute R;
  sr:small absolute R;
  d:DateRec absolute R;
  t:TimeRec absolute R;
 begin with F do case FType of
   CFList:b:=0;
   CFWord:w:=0;
   CFLong:l:=0;
   CFPntr:l:=-1;
   CFReal:rr:=0.0;
   CFStr, CFName:StrRec('',sr,Size);
   CFDate:DateGetD(d);
   CFTime:TimeClear(t);
 end end;

function FldTStr(FT:byte):string;
 begin case FT of
{$IFDEF Cir}
  CFList:FldTStr:=' (byte)';
  CFWord:FldTStr:=' (integer)';
  CFLong:FldTStr:=' (longint)';
  CFPntr:FldTStr:='';
  CFReal:FldTStr:=' (real)';
  CFStr :FldTStr:='';
  CFName:FldTStr:=' ';
  CFDate:FldTStr:='';
  CFTime:FldTStr:='';
{$ENDIF}
{$IFDEF Lat}
  CFList:FldTStr:='counter (byte)';
  CFWord:FldTStr:='counter (integer)';
  CFLong:FldTStr:='counter (longint)';
  CFPntr:FldTStr:='pointer';
  CFReal:FldTStr:='quontity (real)';
  CFStr :FldTStr:='string';
  CFName:FldTStr:='name';
  CFDate:FldTStr:='date';
  CFTime:FldTStr:='time';
{$ENDIF}
 end; end;

function FldTStr0(Fld:TFldRec):string;
 var s:string[15]; s1:string[3];
 begin
  case Fld.FType of
   CFList:s:='byte';
   CFWord:s:='integer';
   CFLong:s:='longint';
   CFPntr:s:='^['+DFArr[Fld.FPtr]^.LName+']';
   CFReal:s:='real';
{$IFDEF Cir}
   CFStr :s:='[';
   CFName:s:='[';
   CFDate:s:='';
   CFTime:s:='';
{$ENDIF}
{$IFDEF Lat}
   CFStr :s:='string[';
   CFName:s:='name[';
   CFDate:s:='date';
   CFTime:s:='time';
{$ENDIF}
  end;
  case Fld.FType of
   CFStr,CFName:begin Str(Fld.Size,s1); s:=s+s1+']' end;
  end;
  FldTStr0:=JustStr(s,15);
 end;

function FieldStr(Fld:TFldRec):string;
 var Opt,ShSz,FStr:String[9]; FW,i:byte;
 begin with Fld do begin
  FStr:=''; FW:=Flags;
  for i:=1 to 6 do begin
   if (FW and 1)>0 then FStr:=FStr+'x' else FStr:=FStr+' ';
   FW:=FW shr 1;
  end;
  Opt:=Blank(3); if FType=CFPntr then Str(FPtr:3,Opt);
  Str(ShowSize:3,ShSZ);
  FieldStr:=
   RecStr(Name,FldMaxLen-1)+VLChar+
   FldTStr0(Fld)+VLChar+
   ShSZ+' '+VLChar+
   FStr+VLChar+
   RecStr(CntName,6);
 end end;

{-----------------------------------------------}

function EditSomePntr(FId:byte; var Pntr:longint):boolean;
 var P:PStdSelectOne;
 begin
  if DFArr[FId]^.Fields^.Length>0 then begin
   New(P,Init(FId,Pntr));
   EditSomePntr:=Desktop^.ExecView(P)<>cmCancel;
   Dispose(P,Done);
  end else ErrNum(esMissFields0)
 end;

{---------------- DataPntr Line: ---------------}

type
 PSInputDPtr = ^TSInputDPtr;
 TSInputDPtr = object (TSInputPntr)
  FId:byte; Ind:longint;
  constructor Init(aLen,aFId:byte);
  procedure HandleEvent(var Event: TEvent); virtual;
  procedure SetData(var Rec); virtual;
  procedure GetData(var Rec); virtual;
  procedure Draw; virtual;
  procedure EditData; virtual;
 end;

constructor TSInputDPtr.Init(aLen,aFId:byte);
 begin inherited Init(aLen); FId:=aFId end;
procedure TSInputDPtr.HandleEvent(var Event: TEvent);
 var PFB:PFileBrouser;
 begin
  inherited HandleEvent(Event);
  if Event.What=evBroadcast then if Event.Command=cmRecErased
   then begin { update erased pointer }
    PFB:=Event.InfoPtr;
    if (Fid=PFB^.MyId) and (Ind=PFB^.LastInd)
     then begin Ind:=-1; DrawView end;
   end;
 end;
procedure TSInputDPtr.SetData(var Rec); begin Ind:=Longint(Rec) end;
procedure TSInputDPtr.GetData(var Rec); begin Longint(Rec):=Ind end;
procedure TSInputDPtr.Draw;
 var ps:PString;
 begin
  New(ps);
  if Ind=-1
{$IFDEF Cir}
   then ps^:='  '+DFArr[FId]^.LName
{$ENDIF}
{$IFDEF Lat}
   then ps^:='nobody from '+DFArr[FId]^.LName
{$ENDIF}
   else ps^:=DFArr[FId]^.StdPtrLine(Ind,4);
  writeStr(0,0,JustStr(ps^,Size.X),1+ord(GetState(sfFocused)));
  Dispose(ps);
 end;
procedure TSInputDPtr.EditData;
 begin EditSomePntr(FId,Ind); DrawView end;

{---------------- Input Layouts: ---------------}

{ $IFDEF Turbo}
const
 R_byte:TRect  = (A:(X:0;Y:0);B:(X: 5;Y:1));
 R_word:TRect  = (A:(X:0;Y:0);B:(X: 8;Y:1));
 R_long:TRect  = (A:(X:0;Y:0);B:(X:14;Y:1));
 R_real:TRect  = (A:(X:0;Y:0);B:(X:14;Y:1));
 R_date:TRect  = (A:(X:0;Y:0);B:(X:10;Y:1));
 R_date10:TRect= (A:(X:0;Y:0);B:(X:12;Y:1));
{ $ENDIF}
{ $IFDEF FPC}
{var
 R_byte:TRect;
 R_word:TRect;
 R_long:TRect;
 R_real:TRect;
 R_date:TRect;
 R_date10:TRect;
}
{ $ENDIF}

constructor TSRecDialog.Init(var Bounds: TRect; ATitle: TTitleStr);
 begin
  inherited Init(Bounds,ATitle);
  HelpCtx:=hcEditRec; VCnt:=0;
 end;
procedure TSRecDialog.AddNewView(ALabel:string; aPV:PView);
 var R:TRect;
 begin
  Carr[VCnt]:=aPV; Insert(aPV);
  aPV^.MoveTo(3+CStrLen(ALabel),2+VCnt);
  R.Assign(1,2+VCnt,CStrLen(ALabel)+2,3+VCnt);
  Insert(New(PLabel,Init(R,ALabel,aPV)));
  Inc(VCnt);
 end;
procedure TSRecDialog.AddNewLine(ALabel:string; aML:byte; cs:boolean);
 var R:TRect;
 begin
  R.Assign(0,0,2+min(aML,Size.X-Length(aLabel)-7),1);
  AddNewView(ALabel,New(PSInputLine,Init(R,aML,cs)));
 end;
procedure TSRecDialog.AddNewStr(ALabel:string; aML:byte; cs:boolean);
 var R:TRect;
 begin
  R.Assign(0,0,2+min(aML,Size.X-Length(aLabel)-7),1);
  AddNewView(ALabel,New(PInputLine0,Init(R,aML,cs)));
 end;
procedure TSRecDialog.AddStrNum(Num:word; aML:byte);
 begin AddNewStr(ErrStr(Num),aML,true) end;
procedure TSRecDialog.AddNewByte(ALabel:string);
 begin AddNewView(ALabel,New(PSInputByte,Init(R_byte))) end;
procedure TSRecDialog.AddBoolNum(Num:word);
 begin
  AddNewView(ErrStr(Num),
   New(PCheckBox,Init(R_byte)))
 end;
procedure TSRecDialog.AddOKCancelButtons;
 Var R:TRect;
 begin
  R.Assign(0, 0, 10, 2);
  bOK:=New(PButton,Init(R,'O~K~',cmOK,bfDefault));
  Insert(bOK); bOK^.MoveTo(Size.X-25,Size.Y-3);
  bCancel:=New(PButton,Init(R,'Cancel',cmCancel,bfNormal));
  Insert(bCancel); bCancel^.MoveTo(Size.X-13,Size.Y-3);
  SelectNext(False);
 end;
procedure TSRecDialog.AddLayoutLine(
   FldT,FSize,SSize,FPtr:byte; var R; Tit:string);
 begin
  case FldT of
   CFList:AddNewByte(Tit);
   CFWord:AddNewView(Tit,New(PSInputWord,Init(R_Word)));
   CFDate:case SSize of
    0,8:AddNewView(Tit,New(PSInputDate  ,Init(R_date)));
    10:AddNewView(Tit,New(PSInputDate10,Init(R_date10)));
   end;
   CFTime:AddNewView(Tit,New(PSInputTime,Init(R_date)));
   CFLong:AddNewView(Tit,New(PSInputLong,Init(R_Long)));
   CFReal:AddNewView(Tit,New(PSInputReal,Init(R_Real)));
   CFStr :AddNewLine(Tit,FSize,True);
   CFName:AddNewLine(Tit,FSize,False);
   CFPntr:AddNewView(Tit,New(PSInputDPtr,Init(Size.X-Length(Tit)-5,FPtr)));
  end;
  Carr[VCnt-1]^.SetData(R);
 end;

procedure TSRecDialog.HandleEvent(var Event: TEvent);
 var Handled:boolean;
 begin
  Handled:=False;
  case Event.What of
   evKeyDown:case Event.KeyCode of
    kbUp:  begin SelectNext(True); Handled:=True end;
    kbEnter:if (Current<>bOK) and (Current<>bCancel) and (VCnt>1)
     then begin SelectNext(False); Handled:=True end;
    kbDown:begin SelectNext(False); Handled:=True end;
    kbF1:  Event.KeyCode:=kbEnter;
   end;
  end;
  if Handled then ClearEvent(Event) else inherited HandleEvent(Event);
 end;

{-----------------------------------------------}

procedure EditSomeString0(Tit:string; var S:string; Len:byte; Cs:boolean);
 var Dialog:PSRecDialog; R: TRect;
 begin
  R.Assign(0, 0, min(80,max(30,Len+Length(Tit)+9)), 7);
  CenterRect(R); New(Dialog, Init(R,''));
  with Dialog^ do begin AddNewStr(Tit,Len,Cs); AddOkCancelButtons end;
  Dialog^.Carr[0]^.SetData(S);
  EditStringOK := DeskTop^.ExecView(Dialog) <> cmCancel;
  if EditStringOK then Dialog^.Carr[0]^.GetData(S);
  Dispose(Dialog, Done);
 end;

procedure EditSomeString(Tit:string; var S:string; Len:byte);
 begin EditSomeString0(Tit,s,Len,True) end;

procedure EditSomeName(Tit:string; var S:string; Len:byte);
 begin EditSomeString0(Tit,s,Len,False) end;

{-----------------------------------------------}

function EditDate(var D:DateRec):boolean;
 var Dialog:PSRecDialog; R: TRect; OK:boolean;
 begin
  R.Assign(0,0,30,7); CenterRect(R);
{$IFDEF Cir}
  New(Dialog, Init(R,''));
  with Dialog^ do begin
   AddNewView(':',New(PSInputDate,Init(R_date)));
{$ENDIF}
{$IFDEF Lat}
  New(Dialog, Init(R,''));
  with Dialog^ do begin
   AddNewView('Date:',New(PSInputDate,Init(R_date)));
{$ENDIF}
   AddOkCancelButtons;
   Carr[0]^.SetData(D);
  end;
  OK := DeskTop^.ExecView(Dialog) <> cmCancel;
  if OK then with Dialog^ do Carr[0]^.GetData(D);
  EditDate:=OK; Dispose(Dialog, Done);
 end;

{-----------------------------------------------}

procedure EditPeriod(var Prd:TPeriod; var OK:boolean);
 var Dialog:PSRecDialog; R: TRect;
 begin
  R.Assign(0,0,30,8); CenterRect(R);
{$IFDEF Cir}
  New(Dialog, Init(R,''));
  with Dialog^ do begin
   AddNewView(':',New(PSInputDate,Init(R_date)));
   AddNewView(':  ',New(PSInputDate,Init(R_date)));
{$ENDIF}
{$IFDEF Lat}
  New(Dialog, Init(R,'Period'));
  with Dialog^ do begin
   AddNewView('Start:',New(PSInputDate,Init(R_date)));
   AddNewView('End:  ',New(PSInputDate,Init(R_date)));
{$ENDIF}
   AddOkCancelButtons;
   Carr[0]^.SetData(Prd.D0);
   Carr[1]^.SetData(Prd.D1);
  end;
  OK := DeskTop^.ExecView(Dialog) <> cmCancel;
  if OK then with Dialog^ do
   begin Carr[0]^.GetData(Prd.D0); Carr[1]^.GetData(Prd.D1) end;
  Dispose(Dialog, Done);
 end;

{-----------------------------------------------}

function InputLayout(var Rec; FId:byte;
 aInd:longint; abwCase,aExFld:byte):boolean;
 var
  Data:Block absolute Rec;
  PDF:PDataFile; s:string[FldMaxLen+1];
  Dialog:PSRecDialog; R:TRect; OK:boolean;
 function GetHSize:byte;
  var i,j:byte; Fld:TFldRec; s:string[FldMaxLen];
  begin
   j:=30;
   with PDF^.Fields^ do for i:=0 to Length-1 do begin
    GetRec(i,Fld); s:=ClearSp(RecStr(Fld.Name,FldMaxLen));
    if i<>aExFld then
     j:=max(j,min(80,6+ord(s[0])+FldStdLinLen(Fld)));
   end;
   GetHSize:=j;
  end;
 procedure AddViews;
  var c:word; i:byte; Fld:TFldRec;
  begin
   c:=1;
   with PDF^.Fields^ do for i:=0 to Length-1 do begin
    GetRec(i,Fld); s:=ClearSp(RecStr(Fld.Name,FldMaxLen))+':';
    if i<>aExFld then with Fld do
     Dialog^.AddLayoutLine(FType,Size,ShowSize,FPtr,Data[c],s);
    inc(c,Fld.Size);
   end;
  end;
 procedure GetRecData;
  var c:word; i,j:byte; Fld:TFldRec;
  begin
   c:=1; j:=0;
   with PDF^.Fields^ do for i:=0 to Length-1 do begin
    GetRec(i,Fld);
    if i<>aExFld then
     begin Dialog^.CArr[j]^.GetData(Data[c]); inc(j) end;
    inc(c,Fld.Size);
   end;
  end;
 function Go:boolean;
  begin
   R.Assign(0,0,GetHSize,PDF^.Fields^.Length+6-ord(abwCase=bwUIncl));
   CenterRect(R);
   Str(aInd,s); New(Dialog,Init(R,PDF^.LName+'['+s+']'));
   AddViews; Dialog^.AddOkCancelButtons;
   OK := DeskTop^.ExecView(Dialog) <> cmCancel;
   if OK then GetRecData;
   Dispose(Dialog, Done);
   Go:=OK;
  end;
 var Ind:longint;
 begin
  PDF:=DFArr[FId]; Ind:=aInd;
  if PDF^.Data^.Length>aInd { Old rec ? }
   then if PDF^.RecProtList.Add(Ind)
    then begin
     InputLayout:=Go; 
     if not PDF^.RecProtList.Del(Ind) then ErrNum(esProtDelErr0) end
    else begin ErrNum(esProtRecordE); InputLayout:=False end
   else InputLayout:=Go;
 end;

{----------------------------------------------------------}

procedure EditFirstFldValue(Fld:TFldRec; var Rec; aCase:byte);
 var aLabel:string[FldMaxLen+1]; R:TRect; Dialog:PSRecDialog;
 begin
  aLabel:=ClearSp(RecStr(Fld.Name,FldMaxLen))+':';
  R.Assign(0,0,max(30,min(80,5+ord(aLabel[0])+FldStdLinLen(Fld))),7);
  CenterRect(R); New(Dialog,Init(R,ErrStr(esInitialVal0+aCase)));
  with Fld do Dialog^.AddLayoutLine(FType,Size,ShowSize,FPtr,Rec,aLabel);
  Dialog^.AddOkCancelButtons;
  if DeskTop^.ExecView(Dialog) <> cmCancel
   then Dialog^.CArr[0]^.GetData(Rec);
  Dispose(Dialog, Done);
 end;

{----------------------------------------------------------}

procedure LepkaNumEdit(FId:byte; var Sors,Dest:longint);
 var R:TRect; Dialog:PSRecDialog;
 begin
  R.Assign(0,0,70,8); CenterRect(R);
  New(Dialog,Init(R,ErrStr(esDefMixRecds)));
{$IFDEF Cir}
  with Dialog^ do begin
   AddLayoutLine(CFPntr,4,50,FId,Sors,' :');
   AddLayoutLine(CFPntr,4,50,FId,Dest,':  ');
{$ENDIF}
{$IFDEF Lat}
  with Dialog^ do begin
   AddLayoutLine(CFPntr,4,50,FId,Sors,'Bad record:');
   AddLayoutLine(CFPntr,4,50,FId,Dest,'Pattern:   ');
{$ENDIF}
   AddOkCancelButtons;
  end;
  if DeskTop^.ExecView(Dialog) <> cmCancel then with Dialog^ do
    begin CArr[0]^.GetData(Sors); CArr[1]^.GetData(Dest) end;
  Dispose(Dialog, Done);
 end;

{----------------------------------------------------------}

begin
{$IFDEF FPC}
{
 with R_byte do begin 
  with A do begin X:=0; Y:=0 end;
 
  with B do begin X:=5; Y:=1 end; end;
 with R_word do begin 
  with A do begin X:=0; Y:=0 end;
  with B do begin X:=8; Y:=1 end; end;
 with R_long do begin 
  with A do begin X:=0; Y:=0 end;
  with B do begin X:=14; Y:=1 end; end;
 with R_real do begin 
  with A do begin X:=0; Y:=0 end;
  with B do begin X:=14; Y:=1 end; end;
 with R_date do begin
  with A do begin X:=0; Y:=0 end;
  with B do begin X:=10; Y:=1 end; end;
 with R_date10 do begin 
  with A do begin X:=0; Y:=0 end;
  with B do begin X:=12; Y:=1 end; end;
}  
{$ENDIF}
end.
