unit tv_box;
interface
{$I options.pp}

uses
 App, Views, Objects, Dialogs, Drivers, Menus;

{------------------------------------------}

const
{$IFDEF Cir}
 VLChar = '';
 RTChar = '';
 URChar = '';
 DTChar = '' {#208};
 SBChar = '' {#209};
{$ENDIF}
{$IFDEF Lat}
 VLChar = '';
 RTChar = '';
 URChar = '';
 SBChar = '';
{$ENDIF}

{----------------------------------------------}

type
 PSMenuBar = ^TSMenuBar;
 TSMenuBar = object (TMenuBar)
  function NewSubView(
   var Bounds: TRect; AMenu: PMenu; AParentMenu: PMenuView): PMenuView; virtual;
 end;
 PSVertSBar = ^TSVertSBar;
 TSVertSBar = object (TView)
  Value:longint;
  constructor Init(aR:TRect);
  procedure Draw; virtual;
  procedure SetValue(aV:longint); virtual;
  procedure SetParams(aV,aMax:longint; aPgStep:byte); virtual;
  function GetPalette: PPalette; virtual;
  procedure HandleEvent(var Event: TEvent); virtual;
 private
  MaxV:longint; YPos,PgStep:byte;
 end;
 PSWindow = ^TSWindow;
 TSWindow = object (TWindow)
  procedure InitFrame; virtual;
 end;
 PSDialog = ^TSDialog;
 TSDialog = object (TDialog)
  procedure InitFrame; virtual;
 end;
 PSApplication = ^TSApplication;
 TSApplication = object (TApplication)
  procedure InitDesktop; virtual;
 end;

{-------------------------------------------------------}


{
 The next few definitions are similar to the original
 definitions from Turbo Vision MsgBox Unit.
 Size of standart Rect for MessageBox is changed,
 and there is a posibility to use different chars for
 frame border.
}

const

{ Message box classes }

  mfWarning      = $0000;       { Display a Warning box }
  mfError        = $0001;       { Dispaly a Error box }
  mfInformation  = $0002;       { Display an Information Box }
  mfConfirmation = $0003;       { Display a Confirmation Box }

{ Message box button flags }

  mfYesButton    = $0100;       { Put a Yes button into the dialog }
  mfNoButton     = $0200;       { Put a No button into the dialog }
  mfOKButton     = $0400;       { Put an OK button into the dialog }
  mfCancelButton = $0800;       { Put a Cancel button into the dialog }

  mfYesNoCancel  = mfYesButton + mfNoButton + mfCancelButton;
                                { Standard Yes, No, Cancel dialog }
  mfOKCancel     = mfOKButton + mfCancelButton;
                                { Standard OK, Cancel dialog }

{ MessageBox displays the given string in a standard sized      }
{ dialog box. Before the dialog is displayed the Msg and Params }
{ are passed to FormatStr.  The resulting string is displayed   }
{ as a TStaticText view in the dialog.                          }

function MessageBox(
 const Msg: String; Params: Pointer; AOptions: Word): Word;
function MessageBoxRect(
 var R: TRect; const Msg: String; Params: Pointer; AOptions: Word): Word;
procedure CenterRect(var R:TRect);

{-------------------------------------------------}

procedure LifeError(s:string);
procedure WarningMsg(s:string);
function  YN_Win2(s1,s2:string):boolean;

{-------------------------------------------------}

implementation

{$IFDEF Turbo}
uses Skl_defs;
{$ENDIF}
{$IFDEF FPC}
uses sk_types;
{$ENDIF}

{---------------------------------------------------}

type
 PSMenuBox = ^TSMenuBox;
 TSMenuBox = object (TMenuBox)
  procedure Draw; virtual;
 end;

procedure TSMenuBox.Draw;
 var i:byte;
 begin
  inherited Draw;
  for i:=1 to Size.Y-2 do begin
   WriteChar(1       ,i,VLChar,1,1);
   WriteChar(Size.X-2,i,VLChar,1,1);
  end;
  WriteChar(Size.X-2,0,URChar,1,1);
 end;

function TSMenuBar.NewSubView(
 var Bounds: TRect; AMenu: PMenu; AParentMenu: PMenuView): PMenuView;
 begin
  NewSubView:=New(PSMenuBox,Init(Bounds,aMenu,aParentMenu));
 end;

{---------------------------------------------------}

constructor TSVertSBar.Init(aR:TRect);
 begin
  inherited Init(aR);
  EventMask := EventMask or ({evBroadcast +} evMouse{ + evKeyboard});
{ Options:=Options or (ofSelectable + ofFirstClick); }
  Value:=0; MaxV:=0; PgStep:=1; YPos:=0;
 end;
function TSVertSBar.GetPalette: PPalette;
 const P: String[Length(CScrollBar)] = CScrollBar;
 begin GetPalette:=@P end;
procedure TSVertSBar.SetValue(aV:longint);
 var NewV:longint;
 begin
  NewV:=max(0,min(MaxV,aV));
  if NewV<>Value then begin
   Value:=NewV; YPos:=1+((Value+1) * (Size.Y-2)-1) div (MaxV+1);
   DrawView; Message(Owner, evBroadcast,cmScrollBarChanged, @Self);
  end
 end;
procedure TSVertSBar.SetParams(aV,aMax:longint; aPgStep:byte);
 begin
  MaxV:=max(0,aMax); PgStep:=aPgStep; SetValue(aV);
 end;
procedure TSVertSBar.Draw;
 var i:byte;
 begin
  if Size.Y>2 then for i:=1 to Size.Y-2 do if i=YPos
   then WriteChar(0,i,'',1,1) else WriteChar(0,i,SBChar,1,1);
  WriteChar(0,0,'',1,1); WriteChar(0,Size.Y-1,'',1,1);
 end;
procedure TSVertSBar.HandleEvent(var Event: TEvent);
 var MFirst, MLast:TPoint;
 begin
  inherited HandleEvent(Event);
  if Event.What=evMouseDown then begin
   MakeLocal(Event.Where,MFirst);
   if YPos<>MFirst.Y
    then if MFirst.Y=0 then SetValue(Value-1)
     else if MFirst.Y=Size.Y-1 then SetValue(Value+1)
      else if MFirst.Y<YPos
       then SetValue(Value-PgStep) else SetValue(Value+PgStep)
    else repeat
     MakeLocal(Event.Where, MLast);
     SetValue((1+(MLast.Y-1)*(MaxV+1)) div (Size.Y-2));
    until not MouseEvent(Event, evMouseMove);
   ClearEvent(Event);
  end;
 end;

{---------------------------------------------------}

type
 PSFrame = ^TSFrame;
 TSFrame = object (TFrame)
  procedure Draw; virtual;
 end;

procedure TSFrame.Draw;
 var i,bs,c:byte; s:TTitleStr; PW:PWindow absolute Owner;
 begin
  inherited Draw;
  if GetState(sfDragging) then c:=5
   else if GetState(sfActive) then c:=3 else c:=1;
  for i:=1 to Size.Y-2 do begin
   WriteChar(0       ,i,VLChar,c,1);
   WriteChar(Size.X-1,i,VLChar,c,1);
  end;
  s:=''; REXpandStr(s,Size.X-2,'');
  WriteStr(0,0,       ''+s+URChar,c);
  WriteStr(0,Size.Y-1,''+s+'',c);
  if Owner<>Nil then begin
   if (PW^.Flags and (wfClose+wfZoom))>0 then bs:=14 else bs:=6;
   if Size.X>bs then begin
    s:=PW^.GetTitle(Size.X-bs); i:=length(s);
    if i>0 then WriteStr((Size.X-i-2) shr 1,0,' '+s+' ',c);
   end;
   if (PW^.Flags and wfClose)>0 then begin
    WriteChar(2,0,'[',c,1);
    WriteChar(3,0,'',5,1);
    WriteChar(4,0,']',c,1);
   end;
  end;
 end;

{----------------------------------------------------}

procedure TSWindow.InitFrame;
 var R:TRect;
 begin
  R.Assign(0,0,Size.X,Size.Y);
  Frame:=New(PSFrame,Init(R));
 end;

{-----------------------------------------------------}

procedure TSDialog.InitFrame;
 var R:TRect;
 begin
  R.Assign(0,0,Size.X,Size.Y);
  Frame:=New(PSFrame,Init(R));
 end;

{------------------------------------------------------}

procedure TSApplication.InitDesktop;
 begin
  inherited InitDesktop;
{$IFDEF Cir}
  Desktop^.BackGround^.Pattern:=DTChar;
{$ENDIF}
 end;

{------------------------------------------------------}

procedure CenterRect(var R:TRect);
 begin
  R.Move((Desktop^.Size.X - R.B.X) div 2, (Desktop^.Size.Y - R.B.Y) div 2);
 end;

function MessageBox(const Msg: String; Params: Pointer;
  AOptions: Word): Word;
 var R: TRect;
 begin
  R.Assign(0, 0, 44, 10);
  CenterRect(R);
  MessageBox := MessageBoxRect(R, Msg, Params, AOptions);
 end;

function MessageBoxRect(var R: TRect; const Msg: String; Params: Pointer;
  AOptions: Word): Word;
 const
  ButtonName: array[0..3] of string[6] =
    ('~Y~es', '~N~o', 'O~K~', 'Cancel');
  Commands: array[0..3] of word =
    (cmYes, cmNo, cmOK, cmCancel);
  Titles: array[0..3] of string[12] =
{$IFDEF Cir}
    ('','','','');
{$ENDIF} {$IFDEF Lat}
    ('Warning','Error','Information','Confirm');
{$ENDIF}
 var
  I, X, ButtonCount: Integer;
  Dialog: PDialog;
  Control: PView;
  ButtonList: array[0..4] of PView;
  S: String;
 begin
  Dialog := New(PSDialog,Init(R, Titles[AOptions and $3]));
  with Dialog^ do begin
   R.Assign(3, 2, Size.X - 2, Size.Y - 3);
   FormatStr(S, Msg, Params^);
   Insert(New(PStaticText, Init(R, S)));
   X := -2; ButtonCount := 0;
   for I := 0 to 3 do if AOptions and ($0100 shl I) <> 0 then begin
    R.Assign(0, 0, 10, 2);
    Control := New(PButton, Init(R, ButtonName[I], Commands[i], bfNormal));
    Inc(X, Control^.Size.X + 2);
    ButtonList[ButtonCount] := Control;
    Inc(ButtonCount);
   end;
   X := (Size.X - X) shr 1;
   for I := 0 to ButtonCount - 1 do begin
    Control := ButtonList[I]; Insert(Control);
    Control^.MoveTo(X, Size.Y - 3);
    Inc(X, Control^.Size.X + 2);
   end;
   SelectNext(False);
  end;
  MessageBoxRect := DeskTop^.ExecView(Dialog);
  Dispose(Dialog, Done);
end;

{---------------------------------------------------}

function YN_Win2(s1,s2:string):boolean;
 begin YN_Win2:=cmOK=MessageBox(
  #3+s1+#13 + #13 + #3+s2+' ?'#13,
  nil, mfConfirmation or mfOKCancel);
 end;
procedure SomeMsg(s:string; mf:word);
 begin
  MessageBox(
    #3+s+' !'#13 + #13 +
{$IFDEF Cir}
    #3' Enter ...'#13,
{$ENDIF}
{$IFDEF Lat}
    #3'Press Enter to continue ...'#13,
{$ENDIF}
    nil, mf or mfOKButton);
 end;
procedure LifeError(s:string); begin SomeMsg(s,mfError) end;
procedure WarningMsg(s:string); begin SomeMsg(s,mfWarning) end;

end.
