unit Unl_Fils;

{*********************************************************************

  (c) 1998 by Skelet in terms of GNU General Public License

  This unit provides special object, called TVirtFile, witch represents
virtual random access file. I'm using this object for DataBase app-s.
  In both OS (DOS, Linux etc.) there is a limit for count of currently
open files (usualy 16). This unit allows unlimited count of virtualy
open files (realy 8 file handlers are used). Cash algorithm closes
file, who is opened ancient if there is no free handlers. If your
application uses more than 8 open files in time critical parts, change
constant MaxFile below (for DOS you may set MaxFile=12, 4 handlers will
remains for text files and other file dependent parts; in Linux maximal
count of open files depends of some kernel constant, in my computer it
is 255, test may be done by program ../test/lim_test). Don't forget 
that 2-3 handlers are used by system - stdin, stdout, stderr.

*********************************************************************}

interface

uses Linux;

type
 PVirtFile = ^TVirtFile;
 TVirtFile = object
  constructor Init(aFName:String);
  procedure FSeek(aPos:LongInt);
  procedure FTrunc(aPos:LongInt);
  procedure BRead (var rec; Size:LongInt);
  procedure BWrite(var rec; Size:LongInt);
  procedure FFlush;
  procedure FClose;
  procedure FOpen;
  function FSize:LongInt;
  function FName:String;
  function FActive:boolean;
  destructor Done;
 private
  NSize:word; PName:^String;
  FInd:byte;  FPos:LongInt;
 end;

var OpenFiles:LongInt;
procedure CloseAllFiles;

implementation

{
uses
 Trace;
}

const
 MaxFile = 7;

type
 FileRec = record
  fdes:LongInt;
  PVF:PVirtFile;
  age:byte;
 end;

var
 FileMap:array[0..MaxFile] of FileRec;

Function Sys_LSeek(F:longint;Off:longint;Whence:longint):longint;
var regs : SysCallregs;
begin
  regs.reg2:=f;
  regs.reg3:=off;
  regs.reg4:=Whence;
  Sys_lseek:=SysCall(SysCall_nr_lseek,regs);
end;
Function Sys_FDataSync(F:longint):boolean;
var regs : SysCallregs;
begin
  regs.reg2:=f;
  Sys_FDataSync:=SysCall(SysCall_nr_fdatasync,regs)=0;
end;

procedure DiskHalt(FName:String; s:String);
 begin
  writeln(StdErr,'Fatal error for file "'+FName+'"');
  writeln(StdErr,'Operation '+s+' failed ...');
  halt;
 end;

procedure RealClose(Ind:byte);
 begin if Ind<>$FF then with FileMap[Ind] do if PVF<>Nil then begin
  if not fdClose(fdes) then DiskHalt(PVF^.PName^,'close');
  age:=MaxFile; PVF^.FInd:=$FF; PVF:=Nil
 end end;

procedure CloseAllFiles;
 var i:byte; begin for i:=0 to MaxFile do RealClose(i) end;

constructor TVirtFile.Init(aFName:String);
 begin
  NSize:=length(aFName)+1; GetMem(PName,NSize);
  Move(aFName,PName^,NSize);
  FPos:=0; FInd:=$FF; Inc(OpenFiles)
 end;
procedure TVirtFile.FSeek(aPos:LongInt); begin FPos:=aPos end;
procedure TVirtFile.FTrunc(aPos:LongInt);
 begin
  FOpen;
  with FileMap[FInd] do
   if not fdTruncate(fdes,aPos) then DiskHalt(PName^,'truncate');
 end;
procedure TVirtFile.BRead(var rec; Size:LongInt);
 begin
  FOpen;
  with FileMap[FInd] do begin
{
   WrLogStr(
    PName^+': BRead at pos $'+HexLong(FPos)+' RecSize=$'+HexLong(Size));
}
   if Sys_LSeek(fdes,FPos,Seek_set)<0
    then DiskHalt(PName^,'seek')
    else if fdRead(fdes,@rec,Size)<>Size
     then DiskHalt(PName^,'BlockRead');
   inc(FPos,Size);
  end;
 end;
procedure TVirtFile.BWrite(var rec; Size:LongInt);
 begin
  FOpen;
  with FileMap[FInd] do begin
   if Sys_LSeek(fdes,FPos,Seek_set)<0
    then DiskHalt(PName^,'seek')
    else if fdWrite(fdes,@rec,Size)<>Size
     then DiskHalt(PName^,'BlockWrite');
   inc(FPos,Size);
  end;
 end;
procedure TVirtFile.FFlush;
 begin
  if FInd<>$FF then with FileMap[FInd] do
   if not Sys_FDataSync(fdes) then DiskHalt(PName^,'SoftFlush')
 end;
procedure TVirtFile.FClose;
 begin RealClose(FInd) end;
procedure TVirtFile.FOpen;
 var i:byte;
 begin if FInd=$FF then begin
  for i:=0 to MaxFile do if MaxFile=FileMap[i].age
   then FInd:=i else inc(FileMap[i].age);
  RealClose(FInd);
  with FileMap[FInd] do begin
   age:=0; PVF:=@Self;
   fdes:=fdOpen(PName^,Open_RdWr+Open_Creat,$1FF);
   if fdes<0 then DiskHalt(PName^,'open');
  end;
 end end;
function TVirtFile.FName:String; begin FName:=PName^ end;
function TVirtFile.FSize:LongInt;
 var res:LongInt;
 begin
  FOpen;
  with FileMap[FInd] do begin
   res:=Sys_LSeek(fdes,0,Seek_end);
   if res<0 then DiskHalt(PName^,'FileSize');
   FSize:=res;
  end;
 end;
function TVirtFile.FActive:boolean; begin FActive:=FInd<>$FF end;
destructor TVirtFile.Done;
 begin FClose; FreeMem(PName,NSize); dec(OpenFiles) end;

var i:byte;
begin
 for i:=0 to MaxFile do with FileMap[i] do
  begin PVF:=Nil; age:=MaxFile end;
 OpenFiles:=0;
end.
